# -*- coding: utf-8 -*-
"""Grid-based simulation of bedrock landslides.

Benjamin Campforts
"""

import numpy as np

from landlab import Component
from landlab.grid.nodestatus import NodeStatus

from ..depression_finder.lake_mapper import _FLOODED
from .cfuncs import _landslide_runout

MAX_HEIGHT_SLOPE = 100  # in m


class BedrockLandslider(Component):
    """
    Landlab component that calculates the location and magnitude of episodic
    bedrock landsliding following the Cullman criterion.
    See the publication:

    Campforts B., Shobe C.M., Steer P., Vanmaercke M., Lague D., Braun J.
    (2020) HyLands 1.0: a hybrid landscape evolution model to simulate the
    impact of landslides and landslide-derived sediment on landscape evolution.
    Geosci Model Dev: 13(9):3863–86.
    `https://dx.doi.org/10.5194/esurf-6-1-2018 <https://dx.doi.org/10.5194/esurf-6-1-2018>`_


    Examples
    --------
    Make a raster model grid and create a plateau

    >>> import numpy as np
    >>> from numpy import testing
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import PriorityFloodFlowRouter, BedrockLandslider

    Make a raster model grid and create a plateau
    * 5x5 grid
    * Initial topography is set to plateau value of 10

    >>> nr = 5
    >>> nc = 5
    >>> dx = 1
    >>> mg = RasterModelGrid((nr, nc), xy_spacing=dx)
    >>> z = mg.add_zeros("topographic__elevation", at="node")
    >>> s = mg.add_zeros("soil__depth", at='node')
    >>> b = mg.add_zeros("bedrock__elevation", at='node')

    make plateau at 10m

    >>> b += 10

    Lower boundary cell to 0

    >>> b[2] = 0
    >>> z[:] = b + s

    Instantiate flow accumulation and BedrockLandslider

    >>> fd = PriorityFloodFlowRouter(
    ...     mg,
    ...     separate_hill_flow=True,
    ...     suppress_out=True,
    ... )
    >>> hy = BedrockLandslider(mg, landslides_return_time=1)

    run flow director and BedrockLandslider for one timestep
    >>> fd.run_one_step()
    >>> vol_SSY ,V_leaving = hy.run_one_step(dt=1)

    After one timestep, we can predict eactly where the landslide will occur.
    The return time is set to 1 year so that probability for sliding is 100%
    The angle of internal driction is 1m/m, the topographical gradient is 10 m/m
    At cardinal cells, the sliding plain will be at (1+10)/2 = 5.5 m/m.
    With a dx of 1, the cardial cell next to the critical sliding node must
    be 5.5 m and the diagonal one at 5.5 * sqrt(2) = 7.8 m

    >>> testing.assert_almost_equal(
    ...     [5.5 * np.sqrt(2), 5.5, 5.5 * np.sqrt(2)], z[6:9], decimal=5
    ... )

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    Campforts B., Shobe C.M., Steer P., Vanmaercke M., Lague D., Braun J.
    (2020) BedrockLandslider 1.0: a hybrid landscape evolution model to simulate the
    impact of landslides and landslide-derived sediment on landscape evolution.
    Geosci Model Dev: 13(9):3863–86.
    `https://dx.doi.org/10.5194/esurf-6-1-2018 <https://dx.doi.org/10.5194/esurf-6-1-2018>`_

    **Additional References**

    None Listed

    """

    _name = "BedrockLandslider"

    _unit_agnostic = True

    _info = {
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "topographic__steepest_slope": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The steepest *downhill* slope",
        },
        "soil__depth": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Depth of soil or weathered bedrock",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "flow__upstream_node_order": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        # Note that this field has to be provided in addition to the \
        # flow__receiver_node and will be used to route sediments over the hillslope
        "hill_flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        # Note that this field has to be provided in addition to the \
        # flow__receiver_proportions and will be used to route sediments
        # over the hillslope
        "hill_flow__receiver_proportions": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of proportion of flow sent to each receiver.",
        },
        "hill_topographic__steepest_slope": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The steepest *downhill* slope",
        },
        "LS_sediment__flux": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m3/s",
            "mapping": "node",
            "doc": "Sediment flux originating from landslides \
                (volume per unit time of sediment entering each node)",
        },
        "landslide__erosion": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Total erosion caused by landsliding ",
        },
        "landslide__deposition": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Total deposition of derived sediment",
        },
        "landslide_sediment_point_source": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m3",
            "mapping": "node",
            "doc": "Landslide derived sediment, as point sources on all the \
                critical nodes where landslides initiate, \
                before landslide runout is calculated ",
        },
    }

    _cite_as = """@Article{gmd-13-3863-2020,
                  AUTHOR = {Campforts B., Shobe C.M., Steer P., Vanmaercke M., Lague D., Braun J.},
                  TITLE = {BedrockLandslider 1.0: a hybrid landscape evolution model to simulate the impact of landslides and landslide-derived sediment on landscape evolution.},
                  JOURNAL = {Geoscientific Model Development},
                  VOLUME = {13},
                  YEAR = {2020},
                  NUMBER = {9},
                  PAGES = {3863--3886},
                  URL = {https://doi.org/10.5194/gmd-13-3863-2020},
                  DOI = {10.5194/gmd-13-3863-2020}
                  }"""

    def __init__(
        self,
        grid,
        angle_int_frict=1.0,
        cohesion_eff=1e4,
        landslides_return_time=1e5,
        rho_r=2700,
        grav=9.81,
        fraction_fines_LS=0,
        phi=0,
        max_pixelsize_landslide=1e9,
        seed=2021,
        verbose_landslides=False,
        landslides_on_boundary_nodes=True,
        critical_sliding_nodes=None,
    ):
        """Initialize the BedrockLandslider model.

        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        angle_int_frict: float, optional
            Materials angle of internal friction in [m/m]
        cohesion_eff : float, optional
            Effective cohesion of material [m L^-1 T^-2].
        landslides_return_time  : float, optional
            Return time for stochastic landslide events to occur
        rho_r : float, optional
            Bulk density rock [m L^-3].
        fraction_fines_LS : float
            Fraction of permanently suspendable fines in bedrock
            Value must be between 0 and 1 [-].
        phi : float, optional
            Sediment porosity, value must be between 0 and 1 [-].
        max_pixelsize_landslide : int , optional
            Maximum size for landslides in number of pixels
        verbose_landslides : bool , optional
            Print output as number of simulated landslides per timestep
        seed : float , optional
            Provide seed to set stochastic model.
            If not provided, seed is set to 2021.
            Provide None to keep current seed.
        landslides_on_boundary_nodes : bool, optional
            Allow landslides to initiate (critical node) and extend over
            boundary nodes.
        critical_sliding_nodes : list, optional
            Provide list with critical nodes where landslides have to initiate
            This cancels the stochastic part of the algorithm and allows the
            user to form landslides at the provided critical nodes.
        """
        super(BedrockLandslider, self).__init__(grid)

        topo = self.grid.at_node["topographic__elevation"]
        soil = self.grid.at_node["soil__depth"]

        if "bedrock__elevation" not in grid.at_node:
            grid.add_field("bedrock__elevation", topo - soil, at="node", dtype=float)

        # Check consistency of bedrock, soil and topographic elevation fields
        if not np.allclose(
            grid.at_node["bedrock__elevation"] + grid.at_node["soil__depth"],
            grid.at_node["topographic__elevation"],
        ):
            raise RuntimeError(
                "The sum of bedrock elevation and topographic elevation should be equal"
            )

        self.initialize_output_fields()

        # Store grid and parameters
        self._angle_int_frict = angle_int_frict
        self._cohesion_eff = cohesion_eff
        self._rho_r = rho_r
        self._grav = grav
        self._fraction_fines_LS = fraction_fines_LS
        self._phi = phi
        self._landslides_return_time = landslides_return_time
        self._max_pixelsize_landslide = max_pixelsize_landslide
        self._verbose_landslides = verbose_landslides
        self._landslides_on_boundary_nodes = landslides_on_boundary_nodes
        self._critical_sliding_nodes = critical_sliding_nodes

        # Data structures to store properties of simulated landslides.
        self._landslides_size = []
        self._landslides_volume = []
        self._landslides_volume_sed = []
        self._landslides_volume_bed = []

        # Check input values
        if phi >= 1.0 or phi < 0.0:
            raise ValueError(f"Porosity must be between 0 and 1 ({phi})")

        if fraction_fines_LS > 1.0 or fraction_fines_LS < 0.0:
            raise ValueError(
                f"Fraction of fines must be between 0 and 1 ({fraction_fines_LS})"
            )

        # Set seed
        if seed is not None:
            np.random.seed(seed)

    # Getters for properties
    @property
    def landslides_size(self):
        """
        List with the size of simulated landslides.
        The list is reset every time the _landslide_erosion function is called
        """
        return self._landslides_size

    @property
    def landslides_volume(self):
        """
        List with the volume of simulated landslides.
        The list is reset every time the _landslide_erosion function is called
        """
        return self._landslides_volume

    @property
    def landslides_volume_sed(self):
        """
        List with the volume of sediment eroded by landslides.
        The list is reset every time the _landslide_erosion function is called
        """
        return self._landslides_volume_sed

    @property
    def landslides_volume_bed(self):
        """
        List with the volume of bedrock eroded by landslides.
        The list is reset every time the _landslide_erosion function is called
        """
        return self._landslides_volume_bed

    def _landslide_erosion(self, dt):
        """
        Calculate bedrock landsliding for a time period 'dt'.

        Parameters
        ----------
        dt: float (time)
            The imposed timestep.

        Returns
        -------
        suspended_sed : float
            Volume of suspended sediment.

        """
        # Pointers
        topo = self.grid.at_node["topographic__elevation"]
        bed = self.grid.at_node["bedrock__elevation"]
        steepest_slope = self.grid.at_node["topographic__steepest_slope"]
        soil_d = self.grid.at_node["soil__depth"]
        landslide_sed_in = self.grid.at_node["landslide_sediment_point_source"]
        landslide__ero = self.grid.at_node["landslide__erosion"]

        # Reset LS Plains
        landslide__ero.fill(0.0)
        # Reset landslide sediment point source field
        landslide_sed_in.fill(0.0)

        # Reset data structures to store properties of simulated landslides.
        self._landslides_size = []
        self._landslides_volume = []
        self._landslides_volume_sed = []
        self._landslides_volume_bed = []

        # Identify flooded nodes
        flood_status = self.grid.at_node["flood_status_code"]
        flooded_nodes = np.nonzero(flood_status == _FLOODED)[0]

        """In the following section the location of critical nodes where
        landsldies are initatated is calcualted, unless these critical nodes
        are provided as critical_sliding_nodes"""
        if self._critical_sliding_nodes is None:
            # Calculate gradients
            height_cell = topo - topo[self.grid.at_node["flow__receiver_node"]]

            height_cell[flooded_nodes] = 0
            height_cell[height_cell > MAX_HEIGHT_SLOPE] = MAX_HEIGHT_SLOPE

            sc_rad = np.arctan(self._angle_int_frict)
            height_critical = np.divide(
                (4 * self._cohesion_eff / (self._grav * self._rho_r))
                * (np.sin(np.arctan(steepest_slope)) * np.cos(sc_rad)),
                1 - np.cos(np.arctan(steepest_slope) - sc_rad),
                where=(1 - np.cos(np.arctan(steepest_slope) - sc_rad)) > 0,
                out=np.zeros_like(steepest_slope),
            )
            spatial_prob = np.divide(
                height_cell,
                height_critical,
                where=height_critical > 0,
                out=np.zeros_like(height_critical),
            )
            spatial_prob[np.arctan(steepest_slope) <= sc_rad] = 0
            spatial_prob[spatial_prob > 1] = 1

            # Temporal probability
            temporal_prob = 1 - np.exp(-dt / self._landslides_return_time)

            # Combined probability
            combined_prob = temporal_prob * spatial_prob
            sliding = np.random.rand(combined_prob.size) < combined_prob

            # Now, find the critical node, which is the receiver of critical_landslide_nodes
            # Critical nodes must be unique (a given node can have more receivers...)
            critical_landslide_nodes = np.unique(
                self.grid.at_node["flow__receiver_node"][np.where(sliding)]
            )
            # Remove boundary nodes
            if not self._landslides_on_boundary_nodes:
                critical_landslide_nodes = critical_landslide_nodes[
                    ~self.grid.node_is_boundary(critical_landslide_nodes)
                ]
        else:
            critical_landslide_nodes = np.array(self._critical_sliding_nodes)

        # output variables
        suspended_sed = 0.0
        if self._verbose_landslides:
            print(f"nbSlides = {len(critical_landslide_nodes)}")

        store_cumul_volume = 0.0
        while critical_landslide_nodes.size > 0:
            crit_node = critical_landslide_nodes[0]  # start at first critical node
            crit_node_el = topo[crit_node]

            # get 8 neighbors and only keep those to active nodes which are upstream
            neighbors = np.concatenate(
                (
                    self.grid.active_adjacent_nodes_at_node[crit_node],
                    self.grid.diagonal_adjacent_nodes_at_node[crit_node],
                )
            )
            neighbors = neighbors[neighbors != -1]
            neighbors_up = neighbors[topo[neighbors] > crit_node_el]

            x_crit_node = self.grid.node_x[crit_node]
            y_crit_node = self.grid.node_y[crit_node]

            dist_to_initial_node = np.sqrt(
                np.add(
                    np.square(x_crit_node - self.grid.node_x[neighbors_up]),
                    np.square(y_crit_node - self.grid.node_y[neighbors_up]),
                )
            )
            slope_neighbors_to_crit_node = (
                topo[neighbors_up] - crit_node_el
            ) / dist_to_initial_node

            neighbors_up = neighbors_up[
                slope_neighbors_to_crit_node > self._angle_int_frict
            ]
            slope_neighbors_to_crit_node = slope_neighbors_to_crit_node[
                slope_neighbors_to_crit_node > self._angle_int_frict
            ]

            if slope_neighbors_to_crit_node.size > 0:
                slope_slide = max(slope_neighbors_to_crit_node)
                store_volume_bed = 0.0
                store_volume_sed = 0.0
                upstream_count = 0
                upstream_neighbors = neighbors_up
                if not self._landslides_on_boundary_nodes:
                    upstream_neighbors = upstream_neighbors[
                        ~self.grid.node_is_boundary(upstream_neighbors)
                    ]
                # Fix sliding angle of particular LS
                sliding_angle = (self._angle_int_frict + slope_slide) / 2.0
                nb_landslide_cells = 0

                # If landslides become unrealistically big, exit algorithm
                while upstream_neighbors.size > 0 and (
                    upstream_count <= self._max_pixelsize_landslide
                    and nb_landslide_cells < 1e5
                ):
                    distance_to_crit_node = np.sqrt(
                        np.add(
                            np.square(
                                x_crit_node - self.grid.node_x[upstream_neighbors[0]]
                            ),
                            np.square(
                                y_crit_node - self.grid.node_y[upstream_neighbors[0]]
                            ),
                        )
                    )
                    new_el = crit_node_el + distance_to_crit_node * sliding_angle
                    nb_landslide_cells += 1
                    if new_el < topo[upstream_neighbors[0]]:
                        # Do actual slide
                        upstream_count += 1
                        sed_landslide_ero = np.clip(
                            min(
                                soil_d[upstream_neighbors[0]],
                                topo[upstream_neighbors[0]] - new_el,
                            ),
                            a_min=0.0,
                            a_max=None,
                        )
                        soil_d[upstream_neighbors[0]] -= sed_landslide_ero
                        topo[upstream_neighbors[0]] = new_el
                        bed_landslide_ero = np.clip(
                            bed[upstream_neighbors[0]]
                            - (new_el - soil_d[upstream_neighbors[0]]),
                            a_min=0.0,
                            a_max=None,
                        )
                        bed[upstream_neighbors[0]] -= bed_landslide_ero
                        topo[upstream_neighbors[0]] = new_el

                        vol_sed = (
                            sed_landslide_ero * (1 - self._phi) * (self.grid.dx ** 2)
                        )
                        vol_bed = bed_landslide_ero * (self.grid.dx ** 2)
                        store_volume_sed = store_volume_sed + vol_sed
                        store_volume_bed = store_volume_bed + vol_bed

                        neighbors = np.concatenate(
                            (
                                self.grid.active_adjacent_nodes_at_node[
                                    upstream_neighbors[0]
                                ],
                                self.grid.diagonal_adjacent_nodes_at_node[
                                    upstream_neighbors[0]
                                ],
                            )
                        )
                        neighbors = neighbors[neighbors != -1]
                        neighbors_up = neighbors[topo[neighbors] > crit_node_el]
                        upstream_neighbors = [*upstream_neighbors, *neighbors_up]

                        temp, idx = np.unique(upstream_neighbors, return_index=True)
                        upstream_neighbors = np.array(upstream_neighbors)
                        upstream_neighbors = upstream_neighbors[np.sort(idx)]
                        if not self._landslides_on_boundary_nodes:
                            upstream_neighbors = upstream_neighbors[
                                ~self.grid.node_is_boundary(upstream_neighbors)
                            ]
                        # if one of the LS pixels also appears in critical_landslide_nodes list,
                        # remove it there so that no new landslide is initialized
                        critical_landslide_nodes = critical_landslide_nodes[
                            np.where(
                                (critical_landslide_nodes != upstream_neighbors[0])
                            )
                        ]

                        landslide__ero[upstream_neighbors[0]] = (
                            sed_landslide_ero + bed_landslide_ero
                        )

                    upstream_neighbors = np.delete(upstream_neighbors, 0, 0)

                store_volume = store_volume_sed + store_volume_bed
                store_cumul_volume += store_volume
                if upstream_count > 0:
                    landslide_sed_in[crit_node] += (store_volume / dt) * (
                        1.0 - self._fraction_fines_LS
                    )
                    suspended_sed += (store_volume / dt) * self._fraction_fines_LS

                    self._landslides_size.append(upstream_count)
                    self._landslides_volume.append(store_volume)
                    self._landslides_volume_sed.append(store_volume_sed)
                    self._landslides_volume_bed.append(store_volume_bed)

            if critical_landslide_nodes.size > 0:
                critical_landslide_nodes = np.delete(critical_landslide_nodes, 0, 0)

        return suspended_sed

    def _landslide_runout(self, dt):
        """
        calculate landslide runout using a non-local deposition algorithm based on:
            * Carretier S., Martinod P., Reich M., Godderis Y. (2016) Modelling
              sediment clasts transport during landscape evolution.
              Earth Surf Dyn: 4(1):237–51.
            * Campforts B., Shobe C.M., Steer P., Vanmaercke M., Lague D., Braun J.
              (2020) HyLands 1.0: a hybrid landscape evolution model to simulate
              the impact of landslides and landslide-derived sediment on landscape
              evolution. Geosci Model Dev: 13(9):3863–86.

        Parameters
        ----------
        dt : float
            timestep.

        Returns
        -------
        dh_hill : float
            Hillslope erosion over the simulated domain.
        V_leaving : float
            Total volume of sediment leaving the simulated domain.
        Qs_coreNodes : float
            Sediment flux over the simulated domain.

        """
        topo = self.grid.at_node["topographic__elevation"]
        bed = self.grid.at_node["bedrock__elevation"]
        soil_d = self.grid.at_node["soil__depth"]
        sed_flux = self.grid.at_node["LS_sediment__flux"]
        stack_rev = np.flip(self.grid.at_node["flow__upstream_node_order"])
        landslide_depo = self.grid.at_node["landslide__deposition"]
        landslide_sed_in = self.grid.at_node["landslide_sediment_point_source"]
        node_status = self.grid.status_at_node

        # Only process core nodes
        stack_rev_sel = stack_rev[node_status[stack_rev] == NodeStatus.CORE]
        receivers = self.grid.at_node["hill_flow__receiver_node"]
        fract_receivers = self.grid.at_node["hill_flow__receiver_proportions"]

        # keep only steepest slope
        slope = np.max(self.grid.at_node["hill_topographic__steepest_slope"], axis=1)
        slope[slope < 0] = 0.0

        flux_in = landslide_sed_in * dt  # flux_in, in m3 per timestep

        # L following carretier 2016
        transport_length_hill = np.matlib.divide(
            self.grid.dx,
            (
                1
                - np.matlib.minimum(
                    np.matlib.square(np.matlib.divide(slope, self._angle_int_frict)),
                    0.999,
                )
            ),
        )
        flux_out = np.zeros(topo.shape)
        dh_hill = np.zeros(topo.shape)
        topo_copy = np.array(topo)
        max_depo = np.zeros(topo.shape)

        _landslide_runout(
            self.grid.dx,
            self._phi,
            stack_rev_sel,
            receivers,
            fract_receivers,
            flux_in,
            transport_length_hill,
            flux_out,
            dh_hill,
            topo_copy,
            max_depo,
        )
        sed_flux[:] = flux_out

        Qs_coreNodes = np.sum(flux_in[self.grid.status_at_node == 0])
        V_leaving = np.sum(flux_in)  # Qs_leaving # in m3 per timestep

        # Change sediment layer
        soil_d[:] += dh_hill
        topo[:] = bed + soil_d

        # Reset Qs
        landslide_sed_in.fill(0.0)
        # Update deposition field
        landslide_depo[:] = dh_hill

        return dh_hill, V_leaving, Qs_coreNodes

    def run_one_step(self, dt):
        """Advance BedrockLandslider component by one time step of size dt.

        Parameters
        ----------
        dt: float (time)
            The imposed timestep.

        Returns
        -------
        vol_SSY : float
            volume of sediment evacuated as syspended sediment.
        V_leaving : float
            Volume of sediment leaving the domain.
        """
        dt = float(dt)

        if self.current_time is None:
            self.current_time = dt
        else:
            self.current_time += dt

        # Landslides
        vol_SSY = self._landslide_erosion(dt)
        dh_hill, V_leaving, Qs_coreNodes = self._landslide_runout(dt)

        return vol_SSY, V_leaving



import numpy as np
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc, free


DTYPE = int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def fill_xy_of_node_hex_horizontal(
    shape,
    np.ndarray[np.double_t, ndim=1] x_of_node,
    np.ndarray[np.double_t, ndim=1] y_of_node,
):
    """Get x and y coordinates for each node."""
    cdef long n_rows = shape[0]
    cdef long n_cols = shape[1]
    cdef long n_nodes = len(x_of_node)
    cdef long longest_row = n_rows // 2
    cdef long row
    cdef long offset
    cdef long *size_of_row
    cdef long *offset_to_row
    cdef double x0

    try:
        size_of_row = <long*>malloc(sizeof(long) * n_rows)
        offset_to_row = <long*>malloc(sizeof(long) * (n_rows + 1))

        size_of_row[0] = shape[1]
        for row in range(1, longest_row + 1):
            size_of_row[row] = size_of_row[row - 1] + 1
        for row in range(longest_row + 1, n_rows):
            size_of_row[row] = size_of_row[row - 1] - 1

        offset_to_row[0] = 0
        for row in range(1, n_rows + 1):
            offset_to_row[row] = offset_to_row[row - 1] + size_of_row[row - 1]

        x0 = longest_row / 2.
        for row in range(longest_row + 1):
            for offset in range(offset_to_row[row], offset_to_row[row + 1]):
                x_of_node[offset] = x0 + offset - offset_to_row[row]
                y_of_node[offset] = row
            x0 -= .5

        x0 += 1.
        for row in range(longest_row + 1, n_rows):
            for offset in range(offset_to_row[row], offset_to_row[row + 1]):
                x_of_node[offset] = x0 + offset - offset_to_row[row]
                y_of_node[offset] = row
            x0 += .5
    finally:
        free(offset_to_row)
        free(size_of_row)


@cython.boundscheck(False)
def fill_xy_of_node_hex_vertical(
    shape,
    np.ndarray[np.double_t, ndim=1] x_of_node,
    np.ndarray[np.double_t, ndim=1] y_of_node,
):
    cdef long n_nodes = len(x_of_node)
    cdef int n_cols = shape[1]
    cdef int longest_column = n_cols // 2
    cdef int size_of_longest_column = longest_column + shape[0]
    cdef int n_rows = 2 * size_of_longest_column - 1
    cdef long size_of_odd_row = n_cols // 2
    cdef long size_of_even_row = n_cols - size_of_odd_row
    cdef long n_middle_rows = 2 * shape[0] - 1
    cdef long n_top_rows = n_cols // 2
    cdef long row
    cdef long n
    cdef long *size_of_row
    cdef long *offset_to_row
    cdef float x0
    cdef long offset

    try:
        size_of_row = <long*>malloc(sizeof(long) * n_rows)
        offset_to_row = <long*>malloc(sizeof(long) * (n_rows + 1))

        # Bottom and top rows
        for row in range(n_top_rows):
            size_of_row[row] = row + 1
            size_of_row[n_rows - row - 1] = row + 1

        # Middle rows
        for row in range(0, n_middle_rows, 2):
            size_of_row[row + n_top_rows] = size_of_even_row
            size_of_row[row + 1 + n_top_rows] = size_of_odd_row

        offset_to_row[0] = 0
        for row in range(1, n_rows + 1):
            offset_to_row[row] = offset_to_row[row - 1] + size_of_row[row - 1]

        for row in range(0, n_rows, 2):
            # offset = offset_to_row[row]
            x0 = - (size_of_row[row] - 1) // 2
            # for n in range(size_of_row[row]):
            #     x_of_node[offset + n] = x0 + n
            #     y_of_node[offset + n] = row
            for offset in range(offset_to_row[row], offset_to_row[row + 1]):
                x_of_node[offset] = x0 + offset - offset_to_row[row]
                y_of_node[offset] = row

        for row in range(1, n_rows, 2):
            # offset = offset_to_row[row]
            x0 = - size_of_row[row] // 2 + .5
            # for n in range(size_of_row[row]):
            #     x_of_node[offset + n] = x0 + n
            #     y_of_node[offset + n] = row
            for offset in range(offset_to_row[row], offset_to_row[row + 1]):
                x_of_node[offset] = x0 + offset - offset_to_row[row]
                y_of_node[offset] = row

        x0 = x_of_node[offset_to_row[n_top_rows]]
        for n in range(n_nodes):
            x_of_node[n] -= x0
    finally:
        free(offset_to_row)
        free(size_of_row)


@cython.boundscheck(False)
def fill_xy_of_node_rect_vertical(
    shape,
    np.ndarray[np.double_t, ndim=1] x_of_node,
    np.ndarray[np.double_t, ndim=1] y_of_node,
):
    """Get x and y coordinates for each node."""
    cdef long n_cols = shape[1]
    cdef long size_of_odd_row = n_cols // 2
    cdef long size_of_even_row = n_cols - size_of_odd_row
    cdef long n_rows = 2 * shape[0]
    cdef long row
    cdef long offset
    cdef long n

    # even rows
    offset = 0
    for row in range(0, n_rows, 2):
        for n in range(size_of_even_row):
            x_of_node[offset + n] = n * 2
            y_of_node[offset + n] = row / 2.
        offset += n_cols

    # odd rows
    offset = size_of_even_row
    for row in range(1, n_rows, 2):
        for n in range(size_of_odd_row):
            x_of_node[offset + n] = 2 * n + 1
            y_of_node[offset + n] = row / 2.
        offset += n_cols


@cython.boundscheck(False)
def fill_xy_of_node_rect_horizontal(
    shape,
    np.ndarray[np.double_t, ndim=1] x_of_node,
    np.ndarray[np.double_t, ndim=1] y_of_node,
):
    """Get x and y coordinates for each node."""
    cdef int n_nodes = x_of_node.shape[0]
    cdef int stride = 2 * shape[1]
    cdef int n_cols = shape[1]
    cdef int row
    cdef int offset
    cdef int n

    for offset in range(0, n_nodes, stride):
        row = (offset // stride) * 2
        for n in range(offset, offset + n_cols):
            x_of_node[n] = n - offset
            y_of_node[n] = row

    for offset in range(shape[1], n_nodes, stride):
        row = (offset // stride) * 2 + 1
        for n in range(offset, offset + n_cols):
            x_of_node[n] = n - offset + .5
            y_of_node[n] = row


@cython.boundscheck(False)
def get_xy_of_node(shape,
                   np.ndarray[np.double_t, ndim=1] x_of_node,
                   np.ndarray[np.double_t, ndim=1] y_of_node):
    """Get x and y coordinates for each node."""
    cdef int n_nodes = x_of_node.shape[0]
    cdef int stride = 2 * shape[1] + 1
    cdef int n_cols = shape[1]
    cdef int row
    cdef int offset
    cdef int n

    for offset in range(0, n_nodes, stride):
        row = (offset // stride) * 2
        for n in range(offset, offset + n_cols):
            x_of_node[n] = n - offset + .5
            y_of_node[n] = row

    for offset in range(shape[1], n_nodes, stride):
        row = (offset // stride) * 2 + 1
        for n in range(offset, offset + n_cols + 1):
            x_of_node[n] = n - offset
            y_of_node[n] = row


@cython.boundscheck(False)
def get_nodes_at_link(shape,
                   np.ndarray[DTYPE_t, ndim=2] nodes_at_link):
    """Get nodes at the tail and head of each node."""
    cdef int n_links = nodes_at_link.shape[0]
    cdef int n_short_rows = (shape[0] + 1) // 2
    cdef int n_long_rows = shape[0] // 2
    cdef int n_cols = shape[1]
    cdef int n_rows = shape[0]
    cdef int stride = 2 * n_cols + 1
    cdef int row
    cdef int n
    cdef int link

    link = 0
    for row in range(n_rows):
        node = row * stride // 2
        if row % 2 == 0:
            for n in range(n_cols - 1):
                nodes_at_link[link, 0] = n + node
                nodes_at_link[link, 1] = n + node + 1
                link += 1
        else:
            for n in range(n_cols):
                nodes_at_link[link, 0] = n + node
                nodes_at_link[link, 1] = n + node + 1
                link += 1

        if row == n_rows - 1:
            break

        if row % 2 == 0:
            for n in range(n_cols):
                nodes_at_link[link, 0] = n + node
                nodes_at_link[link, 1] = n + node + n_cols
                link += 1
                nodes_at_link[link, 0] = n + node
                nodes_at_link[link, 1] = n + node + n_cols + 1
                link += 1
        else:
            nodes_at_link[link, 0] = node
            nodes_at_link[link, 1] = node + n_cols + 1
            link += 1
            for n in range(1, n_cols):
                nodes_at_link[link, 0] = n + node
                nodes_at_link[link, 1] = n + node + n_cols
                link += 1
                nodes_at_link[link, 0] = n + node
                nodes_at_link[link, 1] = n + node + n_cols + 1
                link += 1
            nodes_at_link[link, 0] = node + n_cols
            nodes_at_link[link, 1] = node + n_cols + n_cols
            link += 1


@cython.boundscheck(False)
def get_links_at_patch(shape,
                       np.ndarray[DTYPE_t, ndim=2] links_at_patch):
    """Get links that bound each patch."""
    cdef int n_patches = links_at_patch.shape[0]
    cdef int n_short_rows = (shape[0] + 1) // 2
    cdef int n_long_rows = shape[0] // 2
    cdef int n_cols = shape[1]
    cdef int n_rows = shape[0]
    cdef int row
    cdef int n
    cdef int patch
    cdef int link
    cdef int link_at_row

    patch = 0
    link_at_row = n_cols - 1
    for row in range(n_rows - 1):
        if row % 2 == 0:
            link = link_at_row + 2
            n_patches = n_cols - 1
        else:
            link = link_at_row + 1
            n_patches = n_cols

        # upward pointing
        links_at_patch[patch, 0] = link
        links_at_patch[patch, 1] = link - 1
        links_at_patch[patch, 2] = link - (n_cols + 1)
        patch += 1
        for n in range(n_patches - 1):
            links_at_patch[patch, 0] = links_at_patch[patch - 1, 0] + 2
            links_at_patch[patch, 1] = links_at_patch[patch - 1, 1] + 2
            links_at_patch[patch, 2] = links_at_patch[patch - 1, 2] + 1
            patch += 1

        if row % 2 == 0:
            link = link_at_row
            n_patches = n_cols
        else:
            link = link_at_row + 1
            n_patches = n_cols - 1

        # downward pointing
        if row % 2 == 0:
            links_at_patch[patch, 0] = link + 2 * n_cols
        else:
            links_at_patch[patch, 0] = link + 2 * n_cols - 1
        links_at_patch[patch, 1] = link
        links_at_patch[patch, 2] = link + 1
        patch += 1
        for n in range(n_patches - 1):
            links_at_patch[patch, 0] = links_at_patch[patch - 1, 0] + 1
            links_at_patch[patch, 1] = links_at_patch[patch - 1, 1] + 2
            links_at_patch[patch, 2] = links_at_patch[patch - 1, 2] + 2
            patch += 1

        if row % 2 == 0:
            link_at_row += 3 * n_cols
        else:
            link_at_row += 3 * n_cols - 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
test_user_guide_example.py

This test checks the example presented in the CellLab-CTS User Guide. If this
test fails as part of a compatibility-breaking major version update, the
corresponding code in the User Guide should also be updated. Some limits:
    - plotting is not tested (commented out here)
    - main is renamed here
    - run_duration is reduced to speed up test

Created on Thu Apr 30 11:43:50 2020

@author: gtucker
"""

import time

# import matplotlib (commented out for CI testing)
from numpy import where

from landlab import RasterModelGrid
from landlab.ca.celllab_cts import Transition  # , CAPlotter (# CI testing)
from landlab.ca.raster_cts import RasterCTS


def setup_transition_list():
    """
    Creates and returns a list of Transition() objects to represent state
    transitions for an unbiased random walk.

    Parameters
    ----------
    (none)

    Returns
    -------
    xn_list : list of Transition objects
        List of objects that encode information about the link-state transitions.

    Notes
    -----
    State 0 represents fluid and state 1 represents a particle (such as a
    sediment grain, tea leaf, or solute molecule).

    The states and transitions are as follows:

    Pair state      Transition to       Process             Rate (cells/s)
    ==========      =============       =======             ==============
    0 (0-0)         (none)              -                   -
    1 (0-1)         2 (1-0)             left/down motion    10.0
    2 (1-0)         1 (0-1)             right/up motion     10.0
    3 (1-1)         (none)              -                   -

    """

    # Create an empty transition list
    xn_list = []

    # Append two transitions to the list.
    # Note that the arguments to the Transition() object constructor are:
    #  - Tuple representing starting pair state
    #    (left/bottom cell, right/top cell, orientation)
    #  - Tuple representing new pair state
    #    (left/bottom cell, right/top cell, orientation)
    #  - Transition rate (cells per time step, in this case 1 sec)
    #  - Name for transition
    xn_list.append(Transition((0, 1, 0), (1, 0, 0), 10.0, "left/down motion"))
    xn_list.append(Transition((1, 0, 0), (0, 1, 0), 10.0, "right/up motion"))

    return xn_list


def test_user_guide_example():

    # INITIALIZE

    # User-defined parameters
    nr = 80  # number of rows in grid
    nc = 50  # number of columns in grid
    plot_interval = 0.5  # time interval for plotting, sec
    run_duration = 1.0  # duration of run, sec
    report_interval = 10.0  # report interval, in real-time seconds

    # Remember the clock time, and calculate when we next want to report
    # progress.
    current_real_time = time.time()
    next_report = current_real_time + report_interval

    # Create grid
    mg = RasterModelGrid(shape=(nr, nc), xy_spacing=1.0)

    # Make the boundaries be walls
    mg.set_closed_boundaries_at_grid_edges(True, True, True, True)

    # Create a node-state dictionary
    ns_dict = {0: "fluid", 1: "particle"}

    # Create the transition list
    xn_list = setup_transition_list()

    # Create an array containing the initial node-state values

    # Create the node-state array and attach it to the grid
    node_state_grid = mg.add_zeros("node", "node_state_map", dtype=int)

    # Initialize the node-state array: here, the initial condition is a pile of
    # resting grains at the bottom of a container.
    bottom_rows = where(mg.node_y < 0.1 * nr)[0]
    node_state_grid[bottom_rows] = 1

    # For visual display purposes, set all boundary nodes to fluid
    node_state_grid[mg.closed_boundary_nodes] = 0

    # Create the CA model
    ca = RasterCTS(mg, ns_dict, xn_list, node_state_grid)

    # Set up plotting
    # Set up colors for plotting
    # (commented out for CI testing)
    # grain = '#5F594D'
    # fluid = '#D0E4F2'
    # clist = [fluid,grain]
    # my_cmap = matplotlib.colors.ListedColormap(clist)

    # Create a CAPlotter object for handling screen display
    # (commented out for CI testing)
    # ca_plotter = CAPlotter(ca, cmap=my_cmap)

    # Plot the initial grid (commented out for CI testing)
    # ca_plotter.update_plot()

    # RUN
    current_time = 0.0
    while current_time < run_duration:

        # Once in a while, print out simulation real time to let the user
        # know that the sim is running ok
        current_real_time = time.time()
        if current_real_time >= next_report:
            print(
                "Current simulation time "
                + str(current_time)
                + "  \
                   ("
                + str(int(100 * current_time / run_duration))
                + "%)"
            )
            next_report = current_real_time + report_interval

        # Run the model forward in time until the next output step
        ca.run(current_time + plot_interval, ca.node_state, plot_each_transition=False)
        current_time += plot_interval

        # Plot the current grid (commented out for CI testing)
        # ca_plotter.update_plot()

    # Finalize plot (commented out for CI testing)
    # ca_plotter.finalize()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""ZoneTaxon object of SpeciesEvolver."""
import numpy as np
from pandas import Series

from .base_taxon import Taxon


class ZoneTaxon(Taxon):
    """A taxon based in zones.

    A ``ZoneTaxon`` is composed of members of a lower taxonomic level that each
    exists within a ``Zone`` object. Taxonomic rank is not considered by this
    class despite the use of the term, 'speciation', which is used herein to
    generally describe creation of a child taxon object.

    All zones of the taxon can be obtained with the attribute, ``zones`` that
    are the objects that manage the geographic aspect of taxon member
    populations. The total geographic extent of all populations is depicted by
    the ``range_mask`` attribute. The zones of a ZoneTaxon instance are created
    and updated using a ``ZoneController``. At model time steps, the
    connectivity of zones over time is obtained using attributes of the
    ``Zone`` object.

    The evolution of this taxon type is carried out in two stages during a
    model time step. In the first stage, the zones of the taxon are updated
    as the result of zone connectivity between the prior and current step in
    the method, ``_update_zones``. This method is the primary implementation of
    taxon dispersal and it is called in a stage prior to other evolutionary
    processes so that all taxa are positioned in their landscape locations
    prior to the other processes.

    In the second stage, processes are carried out in methods that are readily
    expanded or overridden when needed. The primary methods of second stage
    macroevolution are ``_evaluate_dispersal``, ``_evaluate_speciation``, and
    ``_evaluate_extinction``. The evaluate dispersal method is intended to
    modify dispersal conducted in the first stage and it has no effect unless
    it is expanded or overridden to have an effect. Processes other than those
    listed above can be called by expanding or overridding the ``_evolve``
    method.

    The taxon is allopatric when it is associated with/exists within multiple
    zones (signifying multiple member populations). A timer is started when a
    taxon becomes allopatric. Allopatric speciation occurs once the timer
    reaches or exceeds the ``time_to_allopatric_speciation`` initialization
    parameter. If the initialization parameter, ``persists_post_speciation``
    is True (default), a child taxon is created in each zone except one zone
    (the largest by area) that becomes the sole zone of the taxon. If
    ``persists_post_speciation`` is set to False, a child taxon is created in
    each and every zone, and the parent no longer occupies any zones, and
    therefore the parent taxon is no longer extant.

    Extinction occurs when the taxon is no longer associated with any zones.
    This occurs when zones in the prior time step do not overlap zones in the
    current time step, signifying the geographic range of the taxon is no more.
    A taxon can become no longer extant also when the taxon speciates and
    ``persists_post_speciation`` is False signifying that the parent taxon
    has evolved into multiple taxon distinct from the original taxon.

    The following columns will be added to the ``record_data_frame`` of the
    SpeciesEvolver instance that tracks objects of this Taxon: 'speciations'
    and 'extinctions', which are the counts of these variables at time steps.
    Another column, 'pseudoextinctions' will be included when
    ``persists_post_speciation`` is False. This variable is the count of
    occurrences when a parent taxon became non-extant due to speciation and not
    because of an absence of zones.
    """

    def __init__(
        self,
        zones,
        parent=None,
        time_to_allopatric_speciation=0,
        persists_post_speciation=True,
    ):
        """Initialize a taxon.

        Parameters
        ----------
        zones : list of Zones
            The initial SpeciesEvolver Zones where the taxon is located.
        parent : Taxon, optional
            A SpeciesEvolver taxon that is the parent taxon. The default value,
            'None' indicates no parent.
        time_to_allopatric_speciation : float, int, optional
            The delay in model time to speciate following taxon geographic
            fragmentation, indicated by multiple objects in the attribute,
            ``zones``. Speciation occurs at the time step when the delay is
            reached or exceeded. The default value of 0 indicates speciation
            occurs at the same time step as geographic fragmentation.
        persists_post_speciation : boolean, optional
            When 'True', the default, taxon persists despite speciation. When
            'False' and following speciation, the taxon is no longer extant.
        """
        super().__init__()

        self.parent = parent
        self._tas = time_to_allopatric_speciation
        self._pps = persists_post_speciation

        # Store zones that each represent an instance of a narrower taxonomic
        # level, e.g. a population.

        self._zones = zones

        # Set taxon time in allopatry.

        self._time_in_allopatry = None
        self._update_allopatry_state()

    @Taxon.extant.setter
    def extant(self, state):
        """Set the living state of the taxon."""
        self._extant = state

        if not state:
            # Ensure the taxon is not associated with zones when it is no
            # longer extant.
            self._zones = []

    @property
    def range_mask(self):
        """A mask representing the geographic extent of the taxon.

        The mask is an array with a length of grid number of nodes. The taxon
        exists at nodes where mask elements are ``True``. The mask of a
        ZoneTaxon object is the union of all of its zone masks.
        """
        masks = [zone.mask for zone in self.zones]
        mask = np.any(masks, 0)

        return mask

    @property
    def zones(self):
        """The zones of the taxon."""
        return self._zones

    def _evolve(self, dt, stage, record):
        """Run a step of evolutionary processes.

        Dispersal resolved during stage 1 can be modified by extending or
        overriding the dispersal evaluation method, as can speciation and
        extinction that are also evaluated in this stage.

        The attribute, ``extant`` is updated by this method.

        Parameters
        ----------
        dt : float
            The model time step duration.
        stage : int
            The evolution stage of the time step.
        record : defaultdict
            The SpeciesEvolver record.

        Returns
        -------
        boolean
           Indicates if the taxon is still evolving. When `False` is returned,
           this method will not be called for the taxon in subsequent stages in
           the current model time step.
        list of Taxon
            The children produced by the taxon at a given stage. The ``evolve``
            method of child taxon will be called in stages following the stage
            the child taxon was produced. An empty list indicates no child
            taxon.
        """
        if stage == 0:
            self._update_zones()
            child_taxa = []

        elif stage == 1:
            # Evaluate macroevolutionary processes now that zones of all taxon
            # objects were updated in stage 1.

            self._evaluate_dispersal(dt)
            self._update_allopatry_state(dt)

            child_taxa = self._evaluate_speciation(dt)
            child_count = len(child_taxa)

            extinct = self._evaluate_extinction(dt)
            pseudoextinct = child_count > 1 and not self._pps
            self.extant = not extinct and not pseudoextinct

            # Update the record.

            record.increment_value("speciations", child_count)
            record.increment_value("extinctions", int(extinct and not pseudoextinct))

            if not self._pps:
                record.increment_value("pseudoextinctions", int(pseudoextinct))

        return stage < 1, child_taxa

    def _update_zones(self):
        """Update the zones of the taxon.

        Dispersal is represented by setting taxon zones to the zones of the
        current time step that overlap the taxon zones of the prior time step
        (`successors of a zone`).
        """
        successors = []

        for zone in self._zones:
            successors.extend(zone.successors)

        self._zones = Series(successors).drop_duplicates().tolist()

    def _update_allopatry_state(self, dt=None):
        """Update taxon time in allopatry.

        Parameter, ``dt`` can optionally be set to increment time in allopatry
        given that the taxon is already allopatric.

        Parameters
        ----------
        dt : float, int, optional
            The model time step duration.
        """
        if len(self.zones) < 2:
            self._time_in_allopatry = None
        elif self._time_in_allopatry is None:
            self._time_in_allopatry = 0
        elif dt is not None:
            self._time_in_allopatry += dt

    def _produce_child_taxon(self, zones):
        """Get the taxon resulting from speciation.

        This method returns a taxon of the same type as the object that
        speciates. This child taxon is initialized with the initialization
        parameter values of the parent object. At minimum in derived
        implementations of this method, the parent taxon of the child should be
        set to `self` to correctly construct the lineage.

        Parameters
        ----------
        zones : list of Zones
            The zones where the child taxon will exist.

        Returns
        -------
        Taxon
            The child taxon.
        """
        taxon_type = type(self)

        child_taxon = taxon_type(
            zones,
            parent=self,
            time_to_allopatric_speciation=self._tas,
            persists_post_speciation=self._pps,
        )

        return child_taxon

    def _evaluate_allopatric_speciation(self, dt):
        """Return child taxa if the taxon is allopatric.

        A child taxon is returned for each zone except the largest zone if
        ``persists_post_speciation`` is True. A child taxon is returned for all
        zones if ``persists_post_speciation`` is False. The ``zones`` attribute
        is set to a list containing the largest zone or no zone when
        ``persists_post_speciation`` is True or False, respectively. This
        method is called by the ``_evaluate_speciation`` method in the second
        stage of taxon evolution.

        Parameter ``dt`` is not used in the ZoneTaxon implementation of this
        method. It is included to follow the pattern of including this
        parameter in ZoneTaxon evolution methods.

        Parameters
        ----------
        dt : float, int
            The model time step duration.

        Returns
        -------
        list of taxon objects
            The taxon objects produced by allopatric speciation. An empty list
            indicates no child objects and no allopatric speciation.
        """
        zones = self.zones
        allopatric = self._time_in_allopatry is not None
        children = []

        if allopatric and self._time_in_allopatry >= self._tas:
            if self._pps:
                # The zone/member with the greatest zone area remains
                # associated with the object.
                idx = np.argmax([zone.area for zone in zones])
                largest_zone = zones.pop(idx)

            for zone in zones:
                child = self._produce_child_taxon([zone])
                children.append(child)

            if self._pps:
                self._zones = [largest_zone]
            else:
                self._zones = []

            self._update_allopatry_state()

        return children

    def _evaluate_dispersal(self, dt):
        """Modify taxon dispersal.

        Population dispersal is principally determined in stage 1 by the
        method, ``_update_zones``. This evaluation method is called by the
        taxon evolve method in stage 2 and allows modification of the stage 1
        dispersal. This method implemented in ZoneTaxon does not modify stage 1
        dispersal. It is intended to be overridden when needed.

        Parameters
        ----------
        dt : float, int
            The model time step duration.
        """
        # pragma: no cover

    def _evaluate_speciation(self, dt):
        """Return child taxa if speciation occurs.

        This method is called by the taxon stage 2 evolve method. The default
        implementation of this method solely gets any taxon objects resulting
        from the method, ``_evaluate_allopatric_speciation``. Other modes of
        speciation, including sympatric, can be evaluated here by expanded this
        ``_evaluate_speciation`` method.

        Parameters
        ----------
        dt : float, int
            The model time step duration.

        Returns
        -------
        list of taxon objects
            The taxon objects produced by allopatric speciation. An empty list
            indicates no child objects and no allopatric speciation.
        """
        child_taxa = self._evaluate_allopatric_speciation(dt)

        return child_taxa

    def _evaluate_extinction(self, dt):
        """Determine if extinction occurs.

        Extinction occurs if no zone/member populations exist. Other conditions
        of extinction can be included by expanding or overridding this method.

        Parameters
        ----------
        dt : float, int
            The model time step duration.

        Returns
        -------
        boolean
            `True` indicates the taxon has become extinct. `False` indicates
            the taxon persists.
        """
        taxon_occupies_no_zones = len(self.zones) == 0

        return taxon_occupies_no_zones



import pytest

from landlab import RasterModelGrid
from landlab.components.plant_competition_ca.plant_competition_ca import VegCA


@pytest.fixture
def ca_veg():
    grid = RasterModelGrid((20, 20), xy_spacing=10e0)
    grid.add_zeros("vegetation__cumulative_water_stress", at="cell")
    grid.add_zeros("vegetation__plant_functional_type", at="cell", dtype=int)
    return VegCA(grid)



import numpy as np
import pandas as pd
import pytest

from landlab import FieldError, RasterModelGrid
from landlab.components import COMPONENTS

_VALID_LOCS = {"grid", "node", "link", "patch", "corner", "face", "cell"}

_REQUIRED_ATTRS = {"doc", "mapping", "dtype", "intent", "optional", "units"}

_EXCLUDE_COMPONENTS = {
    "ChannelProfiler",
    "DrainageDensity",
    "gFlex",
    "HackCalculator",
    "Lithology",
    "LithoLayers",
    "NetworkSedimentTransporter",
    "Profiler",
    "SoilMoisture",
    "Vegetation",
}


@pytest.mark.parametrize("Comp", COMPONENTS)
def test_component_info_unit_agnostic(Comp):
    """Check for a valid _units_agnostic attribute"""
    assert Comp._unit_agnostic in (True, False)


def _add_input_fields_to_grid(cls, grid):
    for name, meta in cls._info.items():
        if meta["intent"].startswith("in"):
            at = cls.var_loc(name)
            dtype = cls.var_type(name)
            if at == "grid":
                grid.at_grid[name] = np.array(0, dtype=dtype)
            else:
                grid.add_zeros(name, at=at, dtype=dtype)

    return grid


@pytest.mark.parametrize("Comp", COMPONENTS)
def test_component_output_fields(Comp):
    """Check that required output fields exist with correct dtypes and locations"""
    if Comp.name in _EXCLUDE_COMPONENTS:
        pytest.skip("component explicitly excluded")

    component_name = Comp._name
    grid = RasterModelGrid((10, 10))

    _add_input_fields_to_grid(Comp, grid)
    Comp(grid)

    for name, meta in Comp._info.items():
        if meta["intent"].endswith("out") and not meta["optional"]:
            at = meta["mapping"]
            if name not in grid[at]:
                raise ValueError(
                    f"{component_name} is missing output variable: {name} at {at}"
                )

            expected_dtype = meta["dtype"]
            actual_dtype = grid[at][name].dtype

            if actual_dtype != expected_dtype:
                raise FieldError(
                    f"{component_name} output required variable: {name} at {at} has "
                    f"incorrect dtype. dtype must be {expected_dtype} and is "
                    f"{actual_dtype}"
                )


@pytest.mark.parametrize("Comp", COMPONENTS)
def test_component_info_missing_attrs(Comp):
    """Check that in/out fields are not missing attributes"""
    component_name = Comp._name

    for name, meta in Comp._info.items():
        at = meta["mapping"]

        missing = ", ".join(sorted(_REQUIRED_ATTRS - set(meta)))
        if missing:
            raise ValueError(
                f"{component_name} is missing attributes ({missing}) about variable: "
                f"{name} at {at}"
            )


@pytest.mark.parametrize("Comp", COMPONENTS)
def test_component_info_unknown_attrs(Comp):
    """Check that in/out fields have valid attributes"""
    component_name = Comp._name

    for name, meta in Comp._info.items():
        at = meta["mapping"]

        unknown = ", ".join(sorted(set(meta) - _REQUIRED_ATTRS))
        if unknown:
            raise ValueError(
                f"{component_name} has extra attributes ({unknown}) about variable: "
                f"{name} at {at}"
            )


@pytest.mark.parametrize("Comp", COMPONENTS)
def test_component_info_valid_dtype(Comp):
    """Check that fields have a valid numpy dtype"""
    component_name = Comp._name

    for name, meta in Comp._info.items():
        dtype = meta["dtype"]
        try:
            np.dtype(dtype)
        except TypeError:
            raise ValueError(
                f"{component_name} has a bad dtype ({dtype}) for variable: {name}"
            )


@pytest.mark.parametrize("Comp", COMPONENTS)
def test_component_info_valid_locations(Comp):
    """Check that fields are defined at valid locations"""
    component_name = Comp._name

    # verify all info exist:
    for name, meta in Comp._info.items():
        at = meta["mapping"]

        # TODO: Verify that all units are UDUNITS compatible.

        if at not in _VALID_LOCS:
            raise ValueError(
                f"{component_name} mapping for variable: {name} is invalid: {at}"
            )


def test_consistent_doc_names():
    out = []
    for comp in COMPONENTS:
        for name in comp._info:
            temp = {"component": comp.__name__, "field": name}
            for key in comp._info[name].keys():
                temp[key] = comp._info[name][key]
            out.append(temp)
    df = pd.DataFrame(out)

    unique_fields = df.field.unique().astype(str)

    bad_fields = {}
    for field in unique_fields:
        where = df.field == field
        if where.sum() > 1:

            sel = df[where]

            doc_vals = df.doc[where].values.astype(str)

            inconsistent = []
            for i in range(len(doc_vals) - 1):
                if doc_vals[i] != doc_vals[-1]:
                    inconsistent.append(sel.component.values[i])

            if len(inconsistent) > 0:
                bad_fields[field] = inconsistent
    if len(bad_fields) > 0:
        msg = "The following fields have inconsistent documentation:\n"
        for field in bad_fields.keys():
            inconsistent = bad_fields[field]
            msg += "\n" + field + ":\n  " + "\n  ".join(inconsistent)

        raise ValueError(msg)



name = "landlab"
cite_as = [
    """@article{hobley2017creative,
    title={Creative computing with Landlab: an open-source toolkit
        for building, coupling, and exploring two-dimensional
        numerical models of Earth-surface dynamics},
    author={Hobley, Daniel EJ and Adams, Jordan M and Nudurupati,
        Sai Siddhartha and Hutton, Eric WH and Gasparini, Nicole M and
        Istanbulluoglu, Erkan and Tucker, Gregory E},
    journal={Earth Surface Dynamics},
    volume={5},
    number={1},
    pages={21},
    year={2017},
    publisher={Copernicus GmbH}
    }
    @article{barnhart2020short,
    author = {Barnhart, K. R. and Hutton, E. W. H. and Tucker, G. E.
        and Gasparini, N. M. and Istanbulluoglu, E. and Hobley,
        D. E. J. and Lyons, N. J. and Mouchene, M. and Nudurupati,
        S. S. and Adams, J. M. and Bandaragoda, C.},
    title = {Short communication: Landlab v2.0: A software package
        for Earth surface dynamics},
    journal = {Earth Surface Dynamics Discussions},
    volume = {2020},
    year = {2020},
    pages = {1--25},
    url = {https://www.earth-surf-dynam-discuss.net/esurf-2020-12/},
    doi = {10.5194/esurf-2020-12}
    }"""
]



#!/usr/env/python

"""
priority_flood_flow_router.py: Component to fill or breach a DEM, accumulate flow and calculate drainage area using the priority flood algorithm.

PriorityFloodFlowRouter is a wrapper of the RichDEM package: https://richdem.readthedocs.io/en/latest/flow_metrics.html

The component merges a filling/breaching algorithm, a flow director as well as a flow accumulator.
Moreover, the component supports the definition of two flow accumulator fields associated to the same grid.
This prevents the user from updating the filling/breaching algorithms in between calculation of flow accumulator one and two.

@author: benjaminCampforts
"""

import copy as cp
from functools import partial

import numpy as np
import numpy.matlib as npm
import richdem as rd

from landlab import Component, FieldError, RasterModelGrid
from landlab.grid.nodestatus import NodeStatus
from landlab.utils.return_array import return_array_at_node

from ...utils.suppress_output import suppress_output
from .cfuncs import _D8_FlowAcc, _D8_flowDir

# Codes for depression status
_UNFLOODED = 0
_PIT = 1
_CURRENT_LAKE = 2
_FLOODED = 3
# Flow metrics resulting in single flow
PSINGLE_FMs = ["D8", "D4", "Rho8", "Rho4"]
# Flow metrics resulting in multiple flow
PMULTIPLE_FMs = ["Quinn", "Freeman", "Holmgren", "Dinf"]


class PriorityFloodFlowRouter(Component):

    """Component to accumulate flow and calculate drainage area based RICHDEM software package.

    See also: https://richdem.readthedocs.io/en/latest/


    .. note::

        The perimeter nodes  NEVER contribute to the accumulating flux, even
        if the  gradients from them point inwards to the main body of the grid.
        This is because under Landlab definitions, perimeter nodes lack cells, so
        cannot accumulate any discharge.

    *FlowAccumulatorPf* stores as *ModelGrid* fields:

    - *'drainage_area'*: Node array of drainage areas
    - *'flood_status_code'*: Map of flood status (_PIT, _CURRENT_LAKE, _UNFLOODED, or _FLOODED).
    - *'surface_water__discharge'*: Node array of discharges.
    - *'Distance to receiver'*: Distance to receiver
    - *'water__unit_flux_in'*: External volume water per area per time input to each node.
    - *'flow__upstream_node_order'*: Node array containing downstream-to-upstream ordered
      list of node IDs.
    - *'flow__receiver_node'*: Node array of receivers (nodes that receive flow), or ITS OWN ID if
      there is no receiver. This array is 2D for *RouteToMany* methods and has the shape
      *(n-nodes x max number of receivers)*.
    - *'flow__receiver_proportions'*: Node array of flow proportions. This
      array is 2D, for *RouteToMany* methods and has the shape
      *(n-nodes x max number of receivers)*.
    - *'topographic__steepest_slope'*: Node array of downhill slopes from each receiver.
      This array is 2D for *RouteToMany* methods and has the shape
      *(n-nodes x max number of receivers)*.
    - *'flow__link_to_receiver_node'*: Node array of links carrying flow.
    - *'flow__receiver_proportion's*: Node array of proportion of flow sent to each receiver.
    - *'depression_free_elevation'*: Depression free land surface topographic elevation, at closed
      borders, value equals -1.

    The following fields are required when an additional hillslope flowrouting
    scheme is required, can be completed with flow acc and discharge if required:

    - *'hill_flow__upstream_node_order'*: Node array containing downstream-to-upstream
      ordered list of node IDs
    - *'hill_flow__receiver_node'*: Node array of receivers (node that receives flow
      from current node)
    - *'hill_topographic__steepest_slope'*: The steepest *downhill* slope.
    - *'hill_flow__receiver_proportions'*: Node array of proportion of flow sent to each
      receiver

    The primary method of this class is :func:`run_one_step`.

    Parameters
    ----------
    grid : ModelGrid
        A Landlab grid.
    surface : str or array_like, optional
        The surface to direct flow across. An at-node field name or an array of length *n_node*.
    flow_metric : str, optional
        String has to be one of 'D8' (O’Callaghan and Mark, 1984), 'Rho8' (Fairfield and Leymarie, 1991),
        'Quinn' (1991), 'Freeman' (1991), 'Holmgren' (1994), 'Dinf' (Tarboton, 1997).
        For details and comparison, see https://richdem.readthedocs.io/en/latest/flow_metrics.html
    runoff_rate : str, array_like, or float, optional
        If provided, sets the runoff rate (m / time) and will be assigned to the grid field
        'water__unit_flux_in'. If a spatially and and temporally variable runoff
        rate is desired, pass this field name and update the field through model
        run time. If both the field and argument are present at the time of
        initialization, runoff_rate will *overwrite* the field. If neither are
        set, defaults to spatially constant unit input.
        Both a runoff_rate array and the 'water__unit_flux_in' field are
        permitted to contain negative values, in which case they mimic
        transmission losses rather than e.g. rain inputs.
    update_flow_depressions : bool, optional
        Build-in depression handler. Can be through filling or breaching (see below).
    update_hill_depressions : bool, optional
        Only needed if DEM needs to be filled separately for second (hill flow) flow accumulator.
        Default behavior is not to execute a separate filling procedure in between the first and
        the second flow accumulator.
    depression_handler : str, optional
        Must be one of 'fill or 'breach'.
        Depression-Filling or breaching algorithm to process depressions

        - 'fill': Depression-Filling.
          Depression-filling is often used to fill in all the depressions
          in a DEM to the level of their lowest outlet or spill-point.
          See also: https://richdem.readthedocs.io/en/latest/depression_filling.html
        - 'breach': Complete Breaching.
          Depression-breaching is used to dig channels from the pit cells
          of a DEM to the nearest cells (in priority-flood sense) outside
          of the depression containing the pit. This resolves the depression
          as all cells in the depression now have a drainage path to the
          edge of the DEM.
          See also: https://richdem.readthedocs.io/en/latest/depression_breaching.html
    exponent : float, optional
        Some methods require an exponent (see flow_metric) Default {1}
    epsilon : bool, optional
        If ``True``, an epsilon gradient is imposed to all flat regions. This ensures
        that there is always a local gradient.
    accumulate_flow : bool, optional
        If ``True`` flow directions and accumulations will be calculated.
        Set to ``False`` when only interested in flow directions
    accumulate_flow_hill : bool, optional
        If ``True`` flow directions and accumulations will be calculated
        for second FD component (Hill). Set to ``False`` when only interested in flow
        directions.
    separate_hill_flow : bool, optional
        For some applications (e.g. *HyLands*) both single and
        multiple flow direction and accumulation is required.
        By calculating them in the same component, we can optimize procedures
        involved with filling and breaching of DEMs
    update_hill_flow_instantaneous : bool, optional
        Update separate hillslope director and accumulator simultaneously on update.
        Set if other operations have to be performed in between updating the
        principle flow properties and the hillslope properties.
    hill_flow_metric : str, optional
        Must be one 'D8' (O’Callaghan and Mark, 1984),'D4' (O’Callaghan and Mark, 1984),
        'Rho8' (Fairfield and Leymarie, 1991), 'Rho4' (Fairfield and Leymarie, 1991),
        'Quinn' (1991) {default},'Freeman' (1991), 'Holmgren' (1994),
        'Dinf' (Tarboton, 1997).
        For details and comparison, see https://richdem.readthedocs.io/en/latest/flow_metrics.html
    hill_exponent : float, optional
        Some methods require an exponent (see flow_metric)
    suppress_out : bool, optional
        Suppress verbose of priority flood algorithm


    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    Barnes, R., 2017. Parallel non-divergent flow accumulation for trillion cell digital elevation models on desktops or clusters. Environmental Modelling & Software 92, 202–212. doi: 10.1016/j.envsoft.2017.02.022

    **Additional References**

    https://richdem.readthedocs.io/en/latest/

    """

    _name = "PriorityFloodFlowRouter"

    _unit_agnostic = True

    _info = {
        "flow__link_to_receiver_node": {
            "dtype": int,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "ID of link downstream of each node, which carries the discharge",
        },
        "drainage_area": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m**2",
            "mapping": "node",
            "doc": "Upstream accumulated surface area contributing to the node's discharge",
        },
        "flood_status_code": {
            "dtype": int,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Map of flood status (_PIT, _CURRENT_LAKE, _UNFLOODED, or _FLOODED).",
        },
        "flow__upstream_node_order": {
            "dtype": int,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "surface_water__discharge": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m**3/s",
            "mapping": "node",
            "doc": "Volumetric discharge of surface water",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "in",
            "optional": True,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "water__unit_flux_in": {
            "dtype": float,
            "intent": "in",
            "optional": True,
            "units": "m/s",
            "mapping": "node",
            "doc": "External volume water per area per time input to each node (e.g., rainfall rate)",
        },
        "topographic__steepest_slope": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The steepest *downhill* slope",
        },
        "squared_length_adjacent": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Length to adjacent nodes, squared (calcualted in advance to save time during calculation",
        },
        "flow__receiver_proportions": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of proportion of flow sent to each receiver.",
        },
        "depression_free_elevation": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Filled land surface topographic elevation, at closed borders, value equals -1!",
        },
        # The following fields are required when an additional
        # hillslope flowrouting scheme is required, can be completed
        # with flow acc and discharge if required
        "hill_drainage_area": {
            "dtype": float,
            "intent": "out",
            "optional": True,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of proportion of flow sent to each receiver.",
        },
        "hill_surface_water__discharge": {
            "dtype": float,
            "intent": "out",
            "optional": True,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of proportion of flow sent to each receiver.",
        },
        "hill_flow__upstream_node_order": {
            "dtype": int,
            "intent": "out",
            "optional": True,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        "hill_flow__receiver_node": {
            "dtype": int,
            "intent": "out",
            "optional": True,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "hill_topographic__steepest_slope": {
            "dtype": float,
            "intent": "out",
            "optional": True,
            "units": "-",
            "mapping": "node",
            "doc": "The steepest *downhill* slope",
        },
        "hill_flow__receiver_proportions": {
            "dtype": float,
            "intent": "out",
            "optional": True,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of proportion of flow sent to each receiver.",
        },
    }

    def __init__(
        self,
        grid,
        surface="topographic__elevation",
        flow_metric="D8",
        runoff_rate=None,
        update_flow_depressions=True,
        depression_handler="fill",
        exponent=1,
        epsilon=True,
        accumulate_flow=True,
        accumulate_flow_hill=False,
        separate_hill_flow=False,
        update_hill_depressions=False,
        update_hill_flow_instantaneous=True,
        hill_flow_metric="Quinn",
        hill_exponent=1,
        suppress_out=True,
    ):
        """Initialize the FlowAccumulator component.

        Saves the grid, tests grid type, tests input types and
        compatibility for the flow_metric and depression_finder
        keyword arguments, tests the argument of runoff, and
        initializes new fields.
        """
        super(PriorityFloodFlowRouter, self).__init__(grid)
        # Keep a local reference to the grid

        self._suppress_output = partial(
            suppress_output, out=suppress_out, err=suppress_out
        )

        # STEP 1: Testing of input values, supplied either in function call or
        # as part of the grid.
        self._test_water_inputs(grid, runoff_rate)

        # Grid type testing
        if not isinstance(self._grid, RasterModelGrid):
            raise FieldError(
                "Flow Accumulator Priority flood only works with regular raster grids, "
                "use default Landlab flow accumulator instead"
            )

        node_cell_area = self._grid.cell_area_at_node.copy()
        node_cell_area[self._grid.closed_boundary_nodes] = 0.0
        self._node_cell_area = node_cell_area
        self._runoff_rate = runoff_rate

        if (flow_metric in PSINGLE_FMs) or (flow_metric in PMULTIPLE_FMs):
            self._flow_metric = flow_metric
        else:
            raise ValueError(
                f"flow metric should be one of these single flow directors : {', '.join(PSINGLE_FMs)} or multiple flow directors: {', '.join(PMULTIPLE_FMs)}"
            )
        if (hill_flow_metric in PSINGLE_FMs) or (hill_flow_metric in PMULTIPLE_FMs):
            self._hill_flow_metric = hill_flow_metric
        else:
            raise ValueError(
                f"flow metric should be one of these single flow directors : {', '.join(PSINGLE_FMs)} or multiple flow directors: {', '.join(PMULTIPLE_FMs)}"
            )

        if depression_handler == "fill":
            self._depression_handler = partial(
                rd.FillDepressions, epsilon=epsilon, in_place=True
            )
        elif depression_handler == "breach":
            self._depression_handler = partial(rd.BreachDepressions, in_place=True)
        else:
            raise ValueError("depression_handler should be one of 'fill' or 'breach'")

        self._exponent = exponent
        self._separate_hill_flow = separate_hill_flow
        self._update_hill_flow_instantaneous = update_hill_flow_instantaneous

        self._update_flow_depressions = update_flow_depressions
        self._update_hill_depressions = update_hill_depressions

        self._hill_exponent = hill_exponent

        if self._separate_hill_flow:
            # Adjust dict
            self._info["hill_drainage_area"]["optional"] = False
            self._info["hill_surface_water__discharge"]["optional"] = False
            self._info["hill_flow__upstream_node_order"]["optional"] = False
            self._info["hill_flow__receiver_node"]["optional"] = False
            self._info["hill_topographic__steepest_slope"]["optional"] = False
            self._info["hill_flow__receiver_proportions"]["optional"] = False
        else:
            self._info["hill_drainage_area"]["optional"] = True
            self._info["hill_surface_water__discharge"]["optional"] = True
            self._info["hill_flow__upstream_node_order"]["optional"] = True
            self._info["hill_flow__receiver_node"]["optional"] = True
            self._info["hill_topographic__steepest_slope"]["optional"] = True
            self._info["hill_flow__receiver_proportions"]["optional"] = True

        self._accumulate_flow = accumulate_flow
        self._accumulate_flow_hill = accumulate_flow_hill

        if not self._accumulate_flow:
            self._info["drainage_area"]["optional"] = True
            self._info["surface_water__discharge"]["optional"] = True
        else:
            self._info["drainage_area"]["optional"] = False
            self._info["surface_water__discharge"]["optional"] = False

        if not self._accumulate_flow_hill:
            self._info["hill_drainage_area"]["optional"] = True
            self._info["hill_surface_water__discharge"]["optional"] = True
        else:

            self._info["hill_drainage_area"]["optional"] = False
            self._info["hill_surface_water__discharge"]["optional"] = False

        self.initialize_output_fields()

        # Make aliases
        if self._accumulate_flow:
            self._drainage_area = self.grid.at_node["drainage_area"]
            self._discharges = self.grid.at_node["surface_water__discharge"]
        self._sort = self.grid.at_node["flow__upstream_node_order"]
        # if multiple flow algorithm is made, the dimensions of the slope and receiver fields change (8 colums for all neightbors)
        if flow_metric in PMULTIPLE_FMs:
            self.grid.at_node["topographic__steepest_slope"] = np.zeros(
                (self.grid.number_of_nodes, 8)
            )
            self.grid.at_node["flow__receiver_node"] = np.zeros(
                (self.grid.number_of_nodes, 8), dtype=int
            )
            self.grid.at_node["flow__receiver_proportions"] = np.zeros(
                (self.grid.number_of_nodes, 8)
            )
            self.grid.at_node["flow__link_to_receiver_node"] = np.zeros(
                (self.grid.number_of_nodes, 8)
            )
        self._slope = self.grid.at_node["topographic__steepest_slope"]
        self._rcvs = self.grid.at_node["flow__receiver_node"]
        self._prps = self.grid.at_node["flow__receiver_proportions"]
        self._recvr_link = self.grid.at_node["flow__link_to_receiver_node"]

        if self._separate_hill_flow:
            if self._accumulate_flow_hill:
                self._hill_drainage_area = self.grid.at_node["hill_drainage_area"]
                self._hill_discharges = self.grid.at_node[
                    "hill_surface_water__discharge"
                ]
            if hill_flow_metric in PMULTIPLE_FMs:
                self.grid.at_node["hill_topographic__steepest_slope"] = np.zeros(
                    (self.grid.number_of_nodes, 8)
                )
                self.grid.at_node["hill_flow__receiver_node"] = np.zeros(
                    (self.grid.number_of_nodes, 8), dtype=int
                )
                self.grid.at_node["hill_flow__receiver_proportions"] = np.zeros(
                    (self.grid.number_of_nodes, 8)
                )
            self._hill_slope = self.grid.at_node["hill_topographic__steepest_slope"]
            self._hill_rcvs = self.grid.at_node["hill_flow__receiver_node"]
            self._hill_prps = self.grid.at_node["hill_flow__receiver_proportions"]

        # Create properties specific to RichDEM
        self._create_richdem_properties()

    @property
    def node_drainage_area(self):
        """Return the drainage area."""
        return self._grid["node"]["drainage_area"]

    @property
    def node_water_discharge(self):
        """Return the surface water discharge."""
        return self._grid["node"]["surface_water__discharge"]

    def _create_richdem_properties(self):
        self._depression_free_dem = cp.deepcopy(
            rd.rdarray(
                self.grid.at_node["topographic__elevation"].reshape(self.grid.shape),
                no_data=-9999,
            )
        )
        self._depression_free_dem.geotransform = [0, 1, 0, 0, 0, -1]

        # Calculate SQUARED length adjacent
        self.grid.at_node["squared_length_adjacent"] = np.concatenate(
            (
                np.ones((self.grid.number_of_nodes, 4)),
                2 * np.ones((self.grid.number_of_nodes, 4)),
            ),
            axis=1,
        )

        self._closed = np.zeros(self._grid.number_of_nodes)
        self._closed[self._grid.status_at_node == NodeStatus.CLOSED] = 1
        self._closed = rd.rdarray(self._closed.reshape(self._grid.shape), no_data=-9999)
        self._closed.geotransform = [0, 1, 0, 0, 0, -1]

    def _test_water_inputs(self, grid, runoff_rate):
        """Test inputs for runoff_rate and water__unit_flux_in."""
        if "water__unit_flux_in" not in grid.at_node:
            if runoff_rate is None:
                # assume that if runoff rate is not supplied, that the value
                # should be set to one everywhere.
                grid.add_ones("water__unit_flux_in", at="node", dtype=float)
            else:
                runoff_rate = return_array_at_node(grid, runoff_rate)
                grid.at_node["water__unit_flux_in"] = runoff_rate
        else:
            if runoff_rate is not None:
                print(
                    "FlowAccumulator found both the field "
                    + "'water__unit_flux_in' and a provided float or "
                    + "array for the runoff_rate argument. THE FIELD IS "
                    + "BEING OVERWRITTEN WITH THE SUPPLIED RUNOFF_RATE!"
                )
                runoff_rate = return_array_at_node(grid, runoff_rate)
                grid.at_node["water__unit_flux_in"] = runoff_rate

    def calc_flow_dir_acc(self, hill_flow=False, update_depressions=True):
        """Calculate flow direction and accumulation using the richdem package"""
        if hill_flow:
            flow_metric = self._hill_flow_metric
        else:
            flow_metric = self._flow_metric

        # 1: Remove depressions
        if update_depressions:
            self.remove_depressions()

        # 2: Flow directions and accumulation
        # D8 flow accumulation in richDEM seems not to differentiate between cardinal and diagonal cells,
        #   so we provide an alternative D8 implementation strategy
        if flow_metric == "D8":
            self._FlowAcc_D8(hill_flow=hill_flow)
        else:

            # Calculate flow direction (proportion) and accumulation using RichDEM
            with self._suppress_output():
                props_Pf = rd.FlowProportions(
                    dem=self._depression_free_dem,
                    method=flow_metric,
                    exponent=self._exponent,
                )

            # Calculate flow accumulation using RichDEM
            if (hill_flow and self._accumulate_flow_hill) or (
                not hill_flow and self._accumulate_flow
            ):
                self._accumulate_flow_RD(props_Pf, hill_flow=hill_flow)

            # Convert flow proportions to landlab structure
            props_Pf = props_Pf.reshape(
                props_Pf.shape[0] * props_Pf.shape[1], props_Pf.shape[2]
            )
            props_Pf_col0 = props_Pf[:, 0]
            props_Pf = np.column_stack(
                (
                    props_Pf[:, 5],
                    props_Pf[:, 7],
                    props_Pf[:, 1],
                    props_Pf[:, 3],
                    props_Pf[:, 6],
                    props_Pf[:, 8],
                    props_Pf[:, 2],
                    props_Pf[:, 4],
                )
            )
            rcvrs = np.concatenate(
                (
                    self._grid.adjacent_nodes_at_node,
                    self._grid.diagonal_adjacent_nodes_at_node,
                ),
                axis=1,
            )
            rcvrs[props_Pf <= 0] = -1
            val = np.arange(0, props_Pf.shape[0])
            rcvrs[props_Pf_col0 == -1, 0] = val[props_Pf_col0 == -1]

            # Links
            recvr_link = np.array(self._grid.d8s_at_node)
            recvr_link[props_Pf <= 0] = -1

            slope_temp = (
                self.grid.at_node["topographic__elevation"].reshape(-1, 1)
                - self.grid.at_node["topographic__elevation"][rcvrs]
            ) / (self.grid.dx * np.sqrt(self.grid.at_node["squared_length_adjacent"]))

            if flow_metric in PSINGLE_FMs:
                slope_temp[rcvrs == -1] = 0
            else:
                props_Pf[props_Pf_col0 == -1, 0] = 1
                props_Pf = props_Pf.astype(np.float64)  # should be float64
                # Now, make sure sum is 1 in 64 bits
                props_Pf[props_Pf == -1] = 0
                rc64_temp = props_Pf / npm.repmat(
                    np.reshape(props_Pf.sum(axis=1), [props_Pf.shape[0], 1]), 1, 8
                )
                props_Pf[props_Pf[:, 0] != 1, :] = rc64_temp[props_Pf[:, 0] != 1, :]
                props_Pf[props_Pf == 0] = -1

            if hill_flow:
                if flow_metric in PSINGLE_FMs:
                    ij_at_max = range(len(rcvrs)), np.argmax(rcvrs, axis=1)
                    self._hill_prps[:] = props_Pf[ij_at_max]
                    self._hill_rcvs[:] = rcvrs[ij_at_max]
                    self._hill_slope[:] = slope_temp[ij_at_max]
                else:
                    self._hill_prps[:] = props_Pf
                    self._hill_rcvs[:] = rcvrs
                    self._hill_slope = slope_temp
                    self._hill_slope[rcvrs == -1] = 0

            else:

                if flow_metric in PSINGLE_FMs:
                    ij_at_max = range(len(rcvrs)), np.argmax(rcvrs, axis=1)
                    self._prps[:] = props_Pf[ij_at_max]
                    self._rcvs[:] = rcvrs[ij_at_max]
                    self._slope[:] = slope_temp[ij_at_max]
                    self._recvr_link[:] = recvr_link[ij_at_max]
                else:
                    self._prps[:] = props_Pf
                    self._rcvs[:] = rcvrs
                    self._slope = slope_temp
                    self._slope[rcvrs == -1] = 0
                    self._recvr_link[:] = recvr_link

    def _FlowAcc_D8(self, hill_flow=False):
        """
        Function to calcualte flow accumulation using the D8 flow algorithm.

        Parameters
        ----------
        hill_flow : Boolean, optional
            Defines which instance of flow accumulation is updated.
            If FALSE, the first, default instance is updated.
            If TRUE, the second, hillslope, instance is updated.
            The default is False.

        Returns
        -------
        None.

        """
        c = self.grid.number_of_node_columns
        dx = self.grid.dx
        activeCells = np.array(
            self._grid.status_at_node != NodeStatus.CLOSED + 0, dtype=int
        )
        receivers = np.array(self.grid.status_at_node, dtype=int)
        distance_receiver = np.zeros((receivers.shape), dtype=float)
        cores = self.grid.core_nodes
        activeCores = cores[activeCells[cores] == 1]
        # Make boundaries to save time with conditionals in c loops
        receivers[np.nonzero(self._grid.status_at_node)] = -1
        steepest_slope = np.zeros((receivers.shape), dtype=float)
        el_dep_free = self._depression_free_dem.reshape(self.grid.number_of_nodes)
        el_ori = self.grid.at_node["topographic__elevation"]
        dist = np.multiply(
            [1, 1, 1, 1, np.sqrt(2), np.sqrt(2), np.sqrt(2), np.sqrt(2)], dx
        )
        ngb = np.zeros((8,), dtype=int)
        el_d = np.zeros((8,), dtype=float)

        # Links
        adj_link = np.array(self._grid.d8s_at_node, dtype=int)
        recvr_link = np.zeros((receivers.shape), dtype=int) - 1

        _D8_flowDir(
            receivers,
            distance_receiver,
            steepest_slope,
            np.array(el_dep_free),
            el_ori,
            dist,
            ngb,
            activeCores,
            activeCells,
            el_d,
            c,
            dx,
            adj_link,
            recvr_link,
        )

        # Calcualte flow acc
        do_FA = False
        if hill_flow:
            if self._accumulate_flow_hill:
                do_FA = True
                a = self._hill_drainage_area
                q = self._hill_discharges
        else:
            if self._accumulate_flow:
                do_FA = True
                a = self._drainage_area
                q = self._discharges

        if do_FA:
            if any(self.grid.at_node["water__unit_flux_in"] != 1):
                wg_q = (
                    self.grid.at_node["water__unit_flux_in"]
                    * self.grid.dx
                    * self.grid.dx
                )
                # Only core nodes (status == 0) need to receive a weight
                wg_q[np.nonzero(self._grid.status_at_node)] = NodeStatus.CORE
                dis = wg_q

            else:
                dis = np.full(self.grid.number_of_nodes, self._node_cell_area)

            da = np.array(self._node_cell_area)
            stack_flip = np.flip(self._sort)
            # Filter out donors giving to receivers being -1
            stack_flip = stack_flip[receivers[stack_flip] != -1]

            _D8_FlowAcc(da, dis, stack_flip, receivers)

            a[:] = da
            q[:] = dis

        # Closed nodes flow to themselves
        val = np.arange(0, receivers.shape[0])
        receivers[receivers == -1] = val[receivers == -1]

        # Restore depression free DEM
        # self._depression_free_dem[self._closed == 1] = -1

        if hill_flow:
            self._hill_prps[:] = 1
            self._hill_rcvs[:] = receivers
            self._hill_slope[:] = steepest_slope
        else:
            self._prps[:] = 1
            self._rcvs[:] = receivers
            self._slope[:] = steepest_slope
            self._recvr_link[:] = recvr_link

    def remove_depressions(self):
        self._depression_free_dem = cp.deepcopy(
            rd.rdarray(
                self.grid.at_node["topographic__elevation"].reshape(self.grid.shape),
                no_data=-9999,
            )
        )
        self._depression_free_dem.geotransform = [0, 1, 0, 0, 0, -1]
        with self._suppress_output():
            self._depression_handler(self._depression_free_dem)
        self._sort[:] = np.argsort(
            np.array(self._depression_free_dem.reshape(self.grid.number_of_nodes))
        )
        self.grid.at_node["depression_free_elevation"] = self._depression_free_dem

    def _accumulate_flow_RD(self, props_Pf, hill_flow=False):
        """
        Function to accumualte flow using the richdem package

        Parameters
        ----------
        props_Pf : float
            flow proportions calcualte with the RichDEM package using the
            FlowProportions function
        hill_flow : Boolean, optional
            Defines which instance of flow accumulation is updated.
            If FALSE, the first, default instance is updated.
            If TRUE, the second, hillslope, instance is updated.
            The default is False.

        Returns
        -------
        None.

        """
        if not hill_flow:
            a = self._drainage_area
            q = self._discharges
        else:
            a = self._hill_drainage_area
            q = self._hill_discharges

        # Create weight for flow accum: both open (status ==1) and closed nodes (status ==4) will have zero weight
        wg = np.full(self.grid.number_of_nodes, self.grid.dx ** 2)

        # Only core nodes (status == 0) need to receive a weight
        wg[self._grid.status_at_node != NodeStatus.CORE] = 0
        wg = rd.rdarray(
            wg.reshape(self.grid.shape),
            no_data=-9999,
        )
        wg.geotransform = [0, 1, 0, 0, 0, -1]

        with self._suppress_output():
            a[:] = np.array(
                rd.FlowAccumFromProps(props=props_Pf, weights=wg).reshape(
                    self.grid.number_of_nodes
                )
            )

        if any(self.grid.at_node["water__unit_flux_in"] != 1):
            wg = self.grid.at_node["water__unit_flux_in"] * self.grid.dx * self.grid.dx
            # Only core nodes (status == 0) need to receive a weight
            wg[self._grid.status_at_node != NodeStatus.CORE] = 0
            wg = rd.rdarray(wg.reshape(self.grid.shape), no_data=-9999)
            wg.geotransform = [0, 1, 0, 0, 0, -1]
            with self._suppress_output():
                q_pf = rd.FlowAccumFromProps(props=props_Pf, weights=wg)
            q[:] = np.array(q_pf.reshape(self.grid.number_of_nodes))
        else:
            q[:] = self._drainage_area

    def update_hill_fdfa(self, update_depressions=False):
        if not self._separate_hill_flow:
            raise ValueError(
                "If hillslope properties are updated, the separate_hill_flow property of the PriorityFloodFlowRouter class should be True upon initialisation"
            )
        self.calc_flow_dir_acc(hill_flow=True, update_depressions=update_depressions)

    def run_one_step(self):
        self.calc_flow_dir_acc(
            hill_flow=False, update_depressions=self._update_flow_depressions
        )
        if self._separate_hill_flow and self._update_hill_flow_instantaneous:
            self.calc_flow_dir_acc(
                hill_flow=True, update_depressions=self._update_hill_depressions
            )



# -*- coding: utf-8 -*-
"""

Unit tests for landlab.data_record.data_record.DataRecord
Test errors/exceptions

Last updated 8/27/2018


"""

import numpy as np
import pytest

from landlab import HexModelGrid, RadialModelGrid, RasterModelGrid, VoronoiDelaunayGrid
from landlab.data_record import DataRecord


def test_misc():
    grid = RasterModelGrid((3, 3))

    # test bad dimension:
    with pytest.raises(ValueError):
        DataRecord(
            grid,
            time=[0.0],
            data_vars={"mean_elev": (["time"], [100.0]), "test": (["bad_dim"], [12])},
        )
    # should return ValueError('Data variable dimensions must be time and/or'
    #                              'item_id')
    # test bad time format:
    with pytest.raises(TypeError):
        DataRecord(grid=grid, time="bad_time")
    # should return TypeError: Time must be a list or an array of length 1

    # test bad datavars format:
    with pytest.raises(TypeError):
        DataRecord(grid, time=[0.0], data_vars=["not a dict"])
    # should return TypeError(('Data variables (data_vars) passed to'
    #                                 ' DataRecord must be a dictionary (see '
    #                                 'documentation for valid structure)'))

    # test bad items format:
    with pytest.raises(TypeError):
        DataRecord(
            grid,
            time=[0.0],
            items=["not a dict"],
            data_vars={"mean_elevation": (["time"], np.array([100]))},
            attrs={"time_units": "y"},
        )
    # Should return TypeError(('You must provide an ''items'' dictionary '
    #                                 '(see documentation for required format)'))
    #
    with pytest.raises(TypeError):
        """Test bad items keys"""
        DataRecord(
            grid,
            time=[0.0],
            items={
                "grid_element": np.array([["node"], ["link"]]),
                "bad_key": np.array([[1], [3]]),
            },
            data_vars={"mean_elevation": (["time"], np.array([100]))},
            attrs={"time_units": "y"},
        )
    # Should return TypeError(('You must provide an ''items'' dictionary '
    #                                  '(see documentation for required format)'))
    with pytest.raises(TypeError):
        """Test bad attrs"""
        DataRecord(
            grid,
            time=[0.0],
            items={
                "grid_element": np.array([["node"], ["link"]]),
                "element_id": np.array([[1], [3]]),
            },
            data_vars={"mean_elevation": (["time"], np.array([100]))},
            attrs=["not a dict"],
        )
    # Should return except AttributeError:
    #                 raise TypeError(('Attributes (attrs) passed to DataRecord'
    #                                 'must be a dictionary'))
    #
    # test bad loc and id:
    rmg = RasterModelGrid((3, 3))
    hmg = HexModelGrid((3, 2), spacing=1.0)
    radmg = RadialModelGrid(n_rings=1, nodes_in_first_ring=5, xy_of_center=(0.0, 0.0))
    vdmg = VoronoiDelaunayGrid(np.random.rand(25), np.random.rand(25))
    my_items_bad_loc = {
        "grid_element": np.array(["node", "bad_loc"]),
        "element_id": np.array([1, 3]),
    }
    my_items_bad_loc2 = {"grid_element": "bad_loc", "element_id": np.array([1, 3])}
    my_items_bad_loc3 = {
        "grid_element": np.array(["node", "node", "node"]),
        "element_id": np.array([1, 3]),
    }
    my_items_bad_id = {
        "grid_element": np.array(["node", "link"]),
        "element_id": np.array([1, 300]),
    }
    my_items_bad_id2 = {
        "grid_element": np.array(["node", "link"]),
        "element_id": np.array([1, -300]),
    }
    my_items_bad_id3 = {
        "grid_element": np.array(["node", "link"]),
        "element_id": np.array([1, 2.0]),
    }
    my_items_bad_id4 = {
        "grid_element": np.array(["node", "link"]),
        "element_id": np.array([1, 2, 3]),
    }
    with pytest.raises(ValueError):
        DataRecord(rmg, items=my_items_bad_loc)
    with pytest.raises(ValueError):
        DataRecord(hmg, items=my_items_bad_loc)
    with pytest.raises(ValueError):
        DataRecord(radmg, items=my_items_bad_loc)
    with pytest.raises(ValueError):
        DataRecord(vdmg, items=my_items_bad_loc)
    # should return ValueError(('One or more of the grid elements provided is/are'
    #                             ' not permitted location for this grid type.'))
    with pytest.raises(ValueError):
        DataRecord(rmg, items=my_items_bad_loc2)
    # should return ValueError: Location provided: bad_loc is not a permitted
    #  location for this grid type.
    with pytest.raises(ValueError):
        DataRecord(rmg, items=my_items_bad_loc3)
    # should return ValueError(('grid_element passed to DataRecord must be '
    #  ' the same length as the number of items or 1.'))
    with pytest.raises(ValueError):
        DataRecord(rmg, items=my_items_bad_id)
    with pytest.raises(ValueError):
        DataRecord(hmg, items=my_items_bad_id)
    with pytest.raises(ValueError):
        DataRecord(radmg, items=my_items_bad_id)
    with pytest.raises(ValueError):
        DataRecord(vdmg, items=my_items_bad_id)
    # should return ValueError(('An item residing at ' + at + ' has an '
    #  'element_id larger than the number of'+ at + 'on the grid.'))
    with pytest.raises(ValueError):
        DataRecord(rmg, items=my_items_bad_id2)
    #  should return ValueError(('An item residing at ' + at + ' has '
    # 'an element id below zero. This is not permitted.'))
    with pytest.raises(ValueError):
        DataRecord(rmg, items=my_items_bad_id3)
    # should return ValueError(('You have passed a non-integer element_id to '
    #                              'DataRecord, this is not permitted.'))
    with pytest.raises(ValueError):
        DataRecord(rmg, items=my_items_bad_id4)
    # should return ValueError(('The number of grid_element passed '
    #  ' to Datarecord must be 1 or equal  to the number of element_id '))


# TIME ONLY
def test_dr_time_bad_add_record(dr_time):
    with pytest.raises(TypeError):
        dr_time.add_record(time="bad_time", new_record=None)
    #  should return TypeError: You have passed a time that is not permitted,
    #  must be list or array
    with pytest.raises(KeyError):
        dr_time.add_record(time=[300.0], item_id=[0], new_record=None)


#  should return KeyError: 'This Datarecord does not hold items'
def test_dr_time_bad_get_data_(dr_time):
    with pytest.raises(KeyError):
        dr_time.get_data(time=[0.0], data_variable="bad_variable")
    # should return KeyError: "the variable 'bad_variable' is not in the Datarecord"
    with pytest.raises(KeyError):
        dr_time.get_data(time=[0.0], item_id=[0], data_variable="mean_elevation")
    # should return KeyError: 'This Datarecord does not hold items'
    with pytest.raises(TypeError):
        dr_time.get_data(time=0.0, data_variable="mean_elevation")


# TypeError('time must be a list or a 1-D array')
def test_dr_time_bad_set_data_(dr_time):
    with pytest.raises(KeyError):
        dr_time.set_data(time=[0.0], data_variable="bad_variable", new_value=105.0)
    # should return KeyError: "the variable 'bad_variable' is not in the Datarecord"
    with pytest.raises(IndexError):
        dr_time.set_data(time=[10.0], data_variable="mean_elevation", new_value=105.0)
    # shoulde return IndexError: 'The time you passed is not currently in the
    # Datarecord, you must the value you pass or first create the new time
    # coordinate using the add_record method.'
    with pytest.raises(KeyError):
        dr_time.set_data(
            time=[0.0], item_id=[0.0], data_variable="mean_elevation", new_value=105.0
        )
    # should return KeyError: 'This datarecord does not hold items'
    with pytest.raises(TypeError):
        dr_time.set_data(time=0.0, data_variable="mean_elevation", new_value=105.0)


# TypeError('time must be a list or a 1-d array')
def test_dr_time_bad_properties(dr_time):
    with pytest.raises(AttributeError):
        dr_time.number_of_items
    with pytest.raises(AttributeError):
        dr_time.item_coordinates


# ITEM ONLY
def test_dr_item_bad_add_record(dr_item):
    with pytest.raises(KeyError):
        dr_item.add_record(
            time=[0.0], item_id=[0], new_record={"new_var": ["new_data"]}
        )
    # should return KeyError: 'This Datarecord does not record time'
    with pytest.raises(ValueError):
        dr_item.add_record(item_id=[10], new_record={"new_var": ["new_data"]})
    # should return ValueError: There is no item with item_id 14, modify
    # the value(s) you pass as item_id or create a new item using the method
    # add_item.
    with pytest.raises(ValueError):
        dr_item.add_record(
            item_id=[0],
            new_item_loc={
                "grid_element": np.array(["cell"]),
                "element_id": np.array([0]),
            },
        )


# should raise ValueError('Use set_data to change the location of'
#                                     ' an item in this DataRecord')
def test_dr_item_bad_add_item(dr_item):
    with pytest.raises(KeyError):
        dr_item.add_item(
            time=[0.0],
            new_item={
                "grid_element": np.array(["node", "link"]),
                "element_id": np.array([1, 5]),
            },
            new_item_spec={"size": (["item_id"], np.array([0.2, 0.3]))},
        )


# should return KeyError: This Datarecord does not record time
def test_dr_item_bad_get_data(dr_item):
    with pytest.raises(KeyError):
        dr_item.get_data(time=[0.0], item_id=[0], data_variable="element_id")
    # should return KeyError: 'This Datarecord does not record time.'
    with pytest.raises(TypeError):
        dr_item.get_data(item_id=0, data_variable="element_id")
    # TypeError('item_id must be a list or a 1-D array')
    with pytest.raises(KeyError):
        dr_item.get_data(item_id=[0], data_variable="bad_variable")


# should return KeyError: "the variable 'bad_variable' is not in the Datarecord"
def test_dr_item_bad_set_data(dr_item):
    with pytest.raises(KeyError):
        dr_item.set_data(
            time=[0.0], item_id=[1], data_variable="element_id", new_value=2
        )
    # should return KeyError: 'This Datarecord does not record time.'
    with pytest.raises(KeyError):
        dr_item.get_data(time=[0.0], item_id=[0], data_variable="bad_variable")
    # should return KeyError: "the variable 'bad_variable' is not in the Datarecord"
    with pytest.raises(IndexError):
        dr_item.set_data(item_id=[3], data_variable="element_id", new_value=2)
    # should return IndexError: The item_id you passed does not exist in this
    # Datarecord
    with pytest.raises(TypeError):
        dr_item.set_data(item_id=3.0, data_variable="element_id", new_value=2)
    # should return TypeError: item_id must be a list or a 1-D array
    with pytest.raises(ValueError):
        dr_item.set_data(item_id=[1], data_variable="element_id", new_value=2.0)
    # ValueError('You have passed a non-integer element_id to DataRecord,
    # this is not permitted')
    with pytest.raises(ValueError):
        dr_item.set_data(item_id=[1], data_variable="element_id", new_value=-2)


# ValueError('You have passed an element id below zero. This is not permitted')
def test_dr_item_bad_properties(dr_item):
    with pytest.raises(AttributeError):
        dr_item.number_of_timesteps
    with pytest.raises(AttributeError):
        dr_item.time_coordinates
    with pytest.raises(AttributeError):
        dr_item.earliest_time
    with pytest.raises(AttributeError):
        dr_item.latest_time
    with pytest.raises(AttributeError):
        dr_item.prior_time


# ITEM AND TIME
def test_dr_2dim_bad_add_record(dr_2dim):
    with pytest.raises(TypeError):
        dr_2dim.add_record(
            time=10.0,
            item_id=[1],
            new_item_loc={
                "grid_element": np.array([["cell"]]),
                "element_id": np.array([[0]]),
            },
        )
    # should return TypeError: You have passed a time that is not permitted,
    # must be list or array.
    with pytest.raises(ValueError):
        dr_2dim.add_record(
            time=[10.0],
            item_id=[10],
            new_item_loc={
                "grid_element": np.array([["cell"]]),
                "element_id": np.array([[0]]),
            },
        )
    # should return ValueError: There is no item with item_id 10, modify the
    # value(s) you pass as item_id or create a new item using the method add_item.
    with pytest.raises(KeyError):
        dr_2dim.add_record(
            time=[10.0],
            item_id=[0],
            new_item_loc={"grid_element": np.array([["cell"]])},
        )
    # should return KeyError: 'You must provide a new_item_loc dictionnary with
    # both grid_element and element_id'
    with pytest.raises(TypeError):
        dr_2dim.add_record(
            time=[10.0],
            item_id=1,
            new_item_loc={
                "grid_element": np.array([["cell"]]),
                "element_id": np.array([[0]]),
            },
        )


# should return except TypeError:
# raise TypeError('item_id must be a list or a 1D array')
def test_dr_2dim_bad_add_item(dr_2dim):
    with pytest.raises(ValueError):
        dr_2dim.add_item(
            new_item={
                "grid_element": np.array([["node"], ["cell"]]),
                "element_id": np.array([[2], [0]]),
            },
            new_item_spec={"size": (["item_id"], [10, 5])},
        )
    # should return ValueError: The items previously defined in this Datarecord have
    # dimensions "time" and "item_id", please provide a "time" for the
    # new item(s)
    with pytest.raises(TypeError):
        dr_2dim.add_item(
            time=[10.0],
            new_item=("not a dict"),
            new_item_spec={"size": (["item_id"], [10, 5])},
        )
    # should return AttributeError raise TypeError: You must provide an new_item
    # dictionary (see documentation for required format)
    with pytest.raises(KeyError):
        dr_2dim.add_item(
            time=[10.0],
            new_item={"grid_element": np.array([["node"], ["cell"]])},
            new_item_spec={"size": (["item_id"], [10, 5])},
        )
    # should return AttributeError: You must provide an new_item dictionnary (see
    # documentation for required format)
    with pytest.raises(TypeError):
        dr_2dim.add_item(
            time=10.0,
            new_item={
                "grid_element": np.array([["node"], ["cell"]]),
                "element_id": np.array([[2], [0]]),
            },
            new_item_spec={"size": (["item_id"], [10, 5])},
        )


# should return TypeError: You have passed a time that is not permitted,
# must be list or array.
def test_dr_2dim_bad_get_data(dr_2dim):
    with pytest.raises(KeyError):
        dr_2dim.get_data(time=[0.0], item_id=1, data_variable="bad_variable")
    with pytest.raises(KeyError):
        dr_2dim.get_data(data_variable="bad_variable")
    # should return KeyError: "the variable 'bad_variable' is not in the Datarecord"
    with pytest.raises(IndexError):
        dr_2dim.get_data(time=[0.0], item_id=[10], data_variable="grid_element")
    # should return IndexError: The item_id you passed does not exist in this
    # Datarecord
    with pytest.raises(TypeError):
        dr_2dim.get_data(time=[0.0], item_id=0, data_variable="element_id")


# TypeError('item_id must be a list or a 1-D array')


def test_dr_2dim_bad_set_data(dr_2dim):
    with pytest.raises(KeyError):
        dr_2dim.set_data(
            time=[0.0], item_id=[1], data_variable="bad_variable", new_value="node"
        )
    with pytest.raises(ValueError):
        dr_2dim.set_data(
            time=[0.0], item_id=[1], data_variable="grid_element", new_value="cell"
        )
    with pytest.raises(IndexError):
        dr_2dim.set_data(
            time=[0.0], item_id=[1.0], data_variable="grid_element", new_value="node"
        )
    with pytest.raises(IndexError):
        dr_2dim.set_data(
            time=[130.0], item_id=[1], data_variable="grid_element", new_value="node"
        )
    with pytest.raises(TypeError):
        dr_2dim.set_data(
            time=[0.0], item_id=1, data_variable="mean_elevation", new_value=105.0
        )


# TypeError('item_id must be a list or a 1-d array')

# NO DIM
def test_dr_nodim_bad_get_data(dr_nodim):
    dr_nodim.add_record(new_record={"mean_elev": (100.0)})
    with pytest.raises(KeyError):
        dr_nodim.get_data(item_id=[0], data_variable="mean_elev")


# should return KeyError: 'This Datarecord does not hold items'



import os

from ..core.messages import deprecation_message


class DeprecatedSignature(DeprecationWarning):

    msg = "You are using a deprecated calling signature."

    def __init__(self, name, old=None, new=None):
        self._name = name
        self._old = old
        self._new = new

        if old:
            self._old = self._construct_call(name, self._old[0], self._old[1])
        if new:
            self._new = self._construct_call(name, self._new[0], self._new[1])

    @staticmethod
    def _construct_call(name, args, kwds):
        signature = ", ".join(
            [repr(arg) for arg in args]
            + ["{k}={v}".format(k=k, v=repr(v)) for k, v in kwds.items()]
        )
        return "{name}({signature})".format(name=name, signature=signature)

    def __str__(self):
        if self._new:
            use = ">>> grid = {call}".format(call=self._new)
        else:
            use = None

        return os.linesep + deprecation_message(self.msg, use=use)



# -*- coding: utf-8 -*-
"""GroundwaterDupuitPercolator Component.

@author: G Tucker, D Litwin, K Barnhart
"""
from warnings import warn

import numpy as np

from landlab import Component
from landlab.grid.mappers import (
    map_mean_of_link_nodes_to_link,
    map_value_at_max_node_to_link,
)
from landlab.utils import return_array_at_link, return_array_at_node


# regularization functions used to deal with numerical demons of seepage
def _regularize_G(u, reg_factor):
    """Smooths transition of step function with an exponential.

    0<=u<=1.
    """
    return np.exp(-(1 - u) / reg_factor)


def _regularize_R(u):
    """ramp function on u."""
    return u * np.greater_equal(u, 0)


def _update_thickness(dt, h0, b, f, dqdx, n, r):
    """analytical solution for the linearized governing equation."""
    out = b * (
        1
        - r
        * np.log(
            1
            + np.exp((1 - (h0 + ((f - dqdx) * dt) / n) / b) / r)
            * (1 - np.exp(-(b - h0) / (b * r)))
        )
    )
    out[f <= dqdx] = (h0 + (1 / n * (f - dqdx)) * dt)[f <= dqdx]
    return out


def get_link_hydraulic_conductivity(grid, K):
    """Returns array of hydraulic conductivity on links, allowing for aquifers
    with laterally anisotropic hydraulic conductivity.

    Parameters
    ----------
    K: (2x2) array of floats (m/s)
        The hydraulic conductivity tensor:
        [[Kxx, Kxy],[Kyx,Kyy]]
    """

    u = grid.unit_vector_at_link
    K_link = np.zeros(len(u))
    for i in range(len(u)):
        K_link[i] = np.dot(np.dot(u[i, :], K), u[i, :])
    return K_link


class GroundwaterDupuitPercolator(Component):
    r"""
    Simulate groundwater flow in a shallow unconfined aquifer.

    The GroundwaterDupuitPercolator solves the Boussinesq equation for
    flow in an unconfined aquifer over an impermeable aquifer base and
    calculates groundwater return flow to the surface. This method uses the
    Dupuit-Forcheimer approximation. This means that the model assumes the
    aquifer is laterally extensive in comparison to its thickness, such that
    the vertical component of flow is negligible. It also assumes that the
    capillary fringe is small, such that the water table can be modeled as a
    free surface. Please consider the applicability of these assumptions when
    using this model. For more details, see component documentation
    :ref:`here <dupuit_theory>`.

    Examples
    --------
    Import the grid class and component

    >>> from landlab import RasterModelGrid
    >>> from landlab.components import GroundwaterDupuitPercolator

    Initialize the grid and component

    >>> grid = RasterModelGrid((10, 10), xy_spacing=10.0)
    >>> elev = grid.add_zeros("topographic__elevation", at="node")
    >>> abe = grid.add_zeros("aquifer_base__elevation", at="node")
    >>> elev[:] = 5.0
    >>> gdp = GroundwaterDupuitPercolator(grid)

    Run component forward.

    >>> dt = 1E4
    >>> for i in range(100):
    ...     gdp.run_one_step(dt)

    When the model generates groundwater return flow, the surface water flux
    out of the domain can be calculated only after a FlowAccumulator is run.
    Below is a more complex example that demonstrates this case.

    >>> from landlab.components import FlowAccumulator

    Set boundary conditions and initialize grid

    >>> grid = RasterModelGrid((5, 41), xy_spacing=10.0)
    >>> grid.set_closed_boundaries_at_grid_edges(True, True, False, True)

    Make a sloping, 3 m thick aquifer, initially fully saturated

    >>> elev = grid.add_zeros("topographic__elevation", at="node")
    >>> elev[:] = grid.x_of_node/100+3
    >>> base = grid.add_zeros("aquifer_base__elevation", at="node")
    >>> base[:] = grid.x_of_node/100
    >>> wt = grid.add_zeros("water_table__elevation", at="node")
    >>> wt[:] = grid.x_of_node/100 + 3

    Initialize components

    >>> gdp = GroundwaterDupuitPercolator(grid, recharge_rate=1E-7)
    >>> fa = FlowAccumulator(grid, runoff_rate='surface_water__specific_discharge')

    Advance timestep. Default units are meters and seconds, though the component
    is unit agnostic.

    >>> dt = 1E3
    >>> for i in range(1000):
    ...     gdp.run_one_step(dt)

    Calculate surface water flux out of domain

    >>> fa.run_one_step()
    >>> np.testing.assert_almost_equal(gdp.calc_sw_flux_out(),0.0005077)


    Notes
    -----
    Below is a summary of the theory and numerical implementation of
    the ``GroundwaterDupuitPercolator``. A complete description can be found
    :ref:`here <dupuit_theory>`.

    Groundwater discharge per unit length, :math:`q`, is calculated as:

    .. math::
        q = -K_{sat} h \big( \nabla z \big) \cos^2 (\alpha)

    where :math:`K_{sat}` is the saturated hydraulic conductivity, :math:`h` is
    the aquifer thickness, and :math:`\alpha` is the slope angle of the aquifer base.

    Surface water discharge per unit area, :math:`q_s`, is calculated as:

    .. math::
        q_s = \mathcal{G}_r \bigg( \frac{h}{d} \bigg) \mathcal{R} \big(-\nabla \cdot q + f \big)

    where :math:`\mathcal{G}_r` is a smoothed step function, :math:`\mathcal{R}` is the ramp function,
    :math:`d` is the regolith thickness, and :math:`f` is the recharge rate.

    The evolution of aquifer thickness is then given by:

    .. math::
        n \frac{\partial h}{\partial t} = f - q_s - \nabla \cdot q

    where :math:`n` is the drainable porosity.

    A semi-analytical approach is used to update aquifer thickness :math:`h`, in which
    the differential equation above is linearized by assuming :math:`\nabla \cdot q` varies
    minimally over the duration of a timestep. In this approach, :math:`q` is
    reported at the beginning of the timestep, while :math:`q_s` is reported at the
    end of the timestep.

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    Litwin, D. G., Tucker, G.E., Barnhart, K. R., Harman, C. J. (2020).
    GroundwaterDupuitPercolator: A Landlab component for groundwater flow.
    Journal of Open Source Software, 5(46), 1935, https://doi.org/10.21105/joss.01935.

    **Additional References**

    Marçais, J., de Dreuzy, J. R. & Erhel, J. Dynamic coupling of subsurface
    and seepage flows solved within a regularized partition formulation.
    Advances in Water Resources 109, 94–105 (2017).

    Childs, E. C. Drainage of Groundwater Resting on a Sloping Bed. Water
    Resources Research 7, 1256–1263 (1971).
    """

    _name = "GroundwaterDupuitPercolator"

    _cite_as = """@article{litwin2020groundwater,
      doi = {10.21105/joss.01935},
      url = {https://doi.org/10.21105/joss.01935},
      year = {2020},
      publisher = {The Open Journal},
      volume = {5},
      number = {46},
      pages = {1935},
      author = {David Litwin and Gregory Tucker and Katherine Barnhart and Ciaran Harman},
      title = {GroundwaterDupuitPercolator: A Landlab component for groundwater flow},
      journal = {Journal of Open Source Software}
    }"""

    _unit_agnostic = True

    _info = {
        "aquifer__thickness": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "thickness of saturated zone",
        },
        "aquifer_base__elevation": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "elevation of impervious layer",
        },
        "aquifer_base__gradient": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/m",
            "mapping": "link",
            "doc": "gradient of the aquifer base in the link direction",
        },
        "average_surface_water__specific_discharge": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/s",
            "mapping": "node",
            "doc": "average surface water specific discharge over variable timesteps",
        },
        "groundwater__specific_discharge": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m2/s",
            "mapping": "link",
            "doc": "discharge per width in link dir",
        },
        "groundwater__velocity": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/s",
            "mapping": "link",
            "doc": "velocity of groundwater in link direction",
        },
        "hydraulic__gradient": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/m",
            "mapping": "link",
            "doc": "gradient of water table in link direction",
        },
        "surface_water__specific_discharge": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/s",
            "mapping": "node",
            "doc": "rate of seepage to surface",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "water_table__elevation": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "elevation of water table",
        },
    }

    def __init__(
        self,
        grid,
        hydraulic_conductivity=0.001,
        porosity=0.2,
        recharge_rate=1.0e-8,
        regularization_f=1e-2,
        courant_coefficient=0.5,
        vn_coefficient=0.8,
        callback_fun=lambda *args, **kwargs: None,
        **callback_kwds,
    ):
        r"""
        Parameters
        ----------
        grid: ModelGrid
            Landlab ModelGrid object
        hydraulic_conductivity: float, field name, array of float or function.
            the aquifer saturated hydraulic conductivity, m/s.
            If function is given, it should take a landlab ModelGrid and return
            an array of floats at link. This may be used if the lateral hydraulic
            conductivity is not vertically homogenous and the effective hydraulic
            conductivity needs to be modified based upon on the position of the
            water table. See component tests for example.
            Default = 0.001 m/s
        porosity: float, field name or array of float
            the drainable porosity of the aquifer [-]
            Default = 0.2
        recharge_rate: float, field name, or array of float
            Rate of recharge, m/s
            Default = 1.0e-8 m/s
        regularization_f: float
            factor controlling the smoothness of the transition between
            surface and subsurface flow
            Default = 0.01
        courant_coefficient: float (-)
            The muliplying factor on the condition that the timestep is
            smaller than the minimum link length over groundwater flow
            velocity. This parameter is only used with
            ``run_with_adaptive_time_step_solver`` and must be greater than
            zero.
            Default = 0.5
        vn_coefficient: float (-)
            The multiplying factor C for the condition :math:`dt >= C*dx^2/(4D)`,
            where :math:`D = Kh/n` is the diffusivity of the Boussinesq
            equation. This arises from a von Neumann stability analysis of
            the Boussinesq equation when the hydraulic gradient is small.
            This parameter is only used with ``run_with_adaptive_time_step_solver``
            and must be greater than zero.
            Default = 0.8
        callback_fun: function(grid, recharge_rate, substep_dt, \*\*kwargs)
            Optional function that will be executed at the end of each sub-timestep
            in the run_with_adaptive_time_step_solver method. Intended purpose
            is to write output not otherwise visible outside of the method call.
            The function should have three required arguments:
            grid: the ModelGrid instance used by GroundwaterDupuitPercolator
            recharge_rate: an array at node that is the specified recharge rate
            substep_dt: the length of the current substep determined internally
            by run_with_adaptive_time_step_solver to meet stability criteria.
        callback_kwds: any additional keyword arguments for the provided callback_fun.
        """
        super().__init__(grid)

        # Shorthand
        self._cores = grid.core_nodes

        # Create fields:

        self._elev = self._grid.at_node["topographic__elevation"]
        self._base = self._grid.at_node["aquifer_base__elevation"]

        self.initialize_output_fields()

        self._wtable = self._grid.at_node["water_table__elevation"]
        self._wtable[grid.closed_boundary_nodes] = 0

        self._thickness = self._grid.at_node["aquifer__thickness"]
        self._thickness[:] = self._wtable - self._base
        self._thickness[grid.closed_boundary_nodes] = 0

        self._hydr_grad = self._grid.at_link["hydraulic__gradient"]
        self._base_grad = self._grid.at_link["aquifer_base__gradient"]

        self._q = self._grid.at_link["groundwater__specific_discharge"]
        self._qs = self._grid.at_node["surface_water__specific_discharge"]
        self._qsavg = self.grid.at_node["average_surface_water__specific_discharge"]

        self._vel = self._grid.at_link["groundwater__velocity"]

        # Convert parameters to fields if needed, and store a reference
        self.K = hydraulic_conductivity
        self.recharge = recharge_rate
        self.n = porosity
        self._r = regularization_f

        # save courant_coefficient (and test)
        self.courant_coefficient = courant_coefficient
        self.vn_coefficient = vn_coefficient

        # set callback function
        self._callback_kwds = callback_kwds
        self.callback_fun = callback_fun

    @property
    def callback_fun(self):
        r"""callback function for adaptive timestep solver

        Parameters
        ----------
        callback_fun: function(grid, recharge_rate, substep_dt, \*\*callback_kwds)
            Optional function that will be executed at the end of each sub-timestep
            in the run_with_adaptive_time_step_solver method. Intended purpose
            is to write output not otherwise visible outside of the method call.
            The function should have three required arguments:
            grid: the ModelGrid instance used by GroundwaterDupuitPercolator
            recharge_rate: an array at node that is the specified recharge rate
            substep_dt: the length of the current substep determined internally
            by run_with_adaptive_time_step_solver to meet stability criteria.
        """
        return self._callback_fun

    @callback_fun.setter
    def callback_fun(self, new_val):
        try:  # New style callback function.
            new_val(self._grid, self.recharge, 0.0, **self._callback_kwds)
            self._callback_fun = new_val
        except TypeError as error:  # Nonfunctional callback function.
            raise ValueError(
                r"%s. Please supply a callback function with the form function(grid, recharge_rate, substep_dt, \*\*kwargs)"
                % error
            )

    @property
    def courant_coefficient(self):
        """Courant coefficient for adaptive time step.

        Parameters
        ----------
        courant_coefficient: float (-)
            The muliplying factor on the condition that the timestep is
            smaller than the minimum link length over groundwater flow
            velocity. This parameter is only used with
            ``run_with_adaptive_time_step_solver`` and must be greater than
            zero.
        """
        return self._courant_coefficient

    @courant_coefficient.setter
    def courant_coefficient(self, new_val):
        if new_val <= 0:
            raise ValueError("courant_coefficient must be > 0.")
        self._courant_coefficient = new_val

    @property
    def vn_coefficient(self):
        """Coefficient for the diffusive timestep condition in
        the adaptive timestep solver.

        Parameters
        ----------
        vn_coefficient: float (-)
            The multiplying factor C for the condition dt >= C*dx^2/(4D),
            where D = Kh/n is the diffusivity of the Boussinesq
            equation. This arises from a von Neumann stability analysis of
            the Boussinesq equation when the hydraulic gradient is small.
            This parameter is only used with ``run_with_adaptive_time_step_solver``
            and must be greater than zero.
        """
        return self._vn_coefficient

    @vn_coefficient.setter
    def vn_coefficient(self, new_val):
        """set coefficient for the diffusive timestep condition in
        the adaptive timestep solver."""
        if new_val <= 0:
            raise ValueError("vn_coefficient must be > 0.")
        self._vn_coefficient = new_val

    @property
    def K(self):
        """hydraulic conductivity at link (m/s)"""
        if self._kfunc:
            self._K = return_array_at_link(self._grid, self._func(self._grid))
        return self._K

    @K.setter
    def K(self, new_val):
        """set hydraulic conductivity at link (m/s)"""
        if callable(new_val):
            self._kfunc = True

            if (
                not isinstance(new_val(self._grid), np.ndarray)
                and len(new_val(self._grid)) == self._grid.number_of_links
            ):
                raise TypeError(
                    """If a function is provided it must take a ModelGrid and return an array of length number_of_links."""
                )
            else:
                self._func = new_val
                self._K = return_array_at_link(self._grid, self._func(self._grid))
        else:
            self._kfunc = False
            self._K = return_array_at_link(self._grid, new_val)

    @property
    def recharge(self):
        """recharge rate (m/s)"""
        return self._recharge

    @recharge.setter
    def recharge(self, new_val):
        """set recharge rate (m/s)"""
        self._recharge = return_array_at_node(self._grid, new_val)

    @property
    def n(self):
        """drainable porosity of the aquifer (-)"""
        return self._n

    @n.setter
    def n(self, new_val):
        """set aquifer drainable porosity (-)"""
        self._n = return_array_at_node(self._grid, new_val)
        self._n_link = map_mean_of_link_nodes_to_link(self._grid, self._n)

    @property
    def number_of_substeps(self):
        """
        The number of substeps used by the run_with_adaptive_time_step_solver
        method in the latest method call.
        """
        if self._num_substeps:
            return self._num_substeps
        else:
            warn("The method run_with_adaptive_time_step_solver has not been used")

        return self._num_substeps

    def calc_recharge_flux_in(self):
        """Calculate flux into the domain from recharge.

        Includes recharge that may immediately become saturation excess
        overland flow. (m3/s)
        """
        return np.sum(
            self._grid.cell_area_at_node[self._cores] * self._recharge[self._cores]
        )

    def calc_gw_flux_out(self):
        """Groundwater flux through open boundaries may be positive (out of the
        domain) or negative (into the domain).

        This function determines the correct sign for specific discharge
        based upon this convention, and sums the flux across the
        boundary faces. (m3/s)
        """
        # get links at open boundary nodes
        open_nodes = self._grid.open_boundary_nodes
        links_at_open = self._grid.links_at_node[open_nodes]
        link_dirs_at_open = self._grid.active_link_dirs_at_node[open_nodes]

        # find active links at open boundary nodes
        active_links_at_open = links_at_open[link_dirs_at_open != 0]
        active_link_dirs_at_open = link_dirs_at_open[link_dirs_at_open != 0]

        # get groundwater specific discharge at these locations
        q_at_open_links = self._grid.at_link["groundwater__specific_discharge"][
            active_links_at_open
        ]

        # get cell widths at these locations
        faces = self._grid.face_at_link[active_links_at_open]
        face_widths = self._grid.length_of_face[faces]

        # get volume flux out at these locations
        gw_volume_flux_rate_out = (
            q_at_open_links * active_link_dirs_at_open * face_widths
        )

        return np.sum(gw_volume_flux_rate_out)

    def calc_sw_flux_out(self):
        """Surface water flux out of the domain through seepage and saturation
        excess.

        Note that model does not allow for reinfiltration.  (m3/s)
        """
        return np.sum(
            self._grid.at_node["surface_water__discharge"][
                self._grid.open_boundary_nodes
            ]
        )

    def calc_gw_flux_at_node(self):
        """Calculate the sum of the groundwater flux leaving a node.

        (m2/s)
        """
        gw = (
            self._grid.at_link["groundwater__specific_discharge"][
                self._grid.links_at_node
            ]
            * self._grid.link_dirs_at_node
        )
        gw_out = -np.sum(gw * (gw < 0), axis=1)
        return gw_out

    def calc_total_storage(self):
        """calculate the current water storage in the aquifer (m3)"""
        return np.sum(
            self._n[self._cores]
            * self._grid.cell_area_at_node[self._cores]
            * self._grid.at_node["aquifer__thickness"][self._cores]
        )

    def run_one_step(self, dt):
        """Advance component by one time step of size dt.

        Parameters
        ----------
        dt: float
            The imposed timestep.
        """

        # check water table above surface
        if (self._wtable > self._elev).any():
            warn(
                "water table above elevation surface. "
                "Setting water table elevation here to "
                "elevation surface"
            )
            self._wtable[self._wtable > self._elev] = self._elev[
                self._wtable > self._elev
            ]
            self._thickness[self._cores] = (self._wtable - self._base)[self._cores]

        # Calculate base gradient
        self._base_grad[self._grid.active_links] = self._grid.calc_grad_at_link(
            self._base
        )[self._grid.active_links]
        cosa = np.cos(np.arctan(self._base_grad))

        # Calculate hydraulic gradient
        self._hydr_grad[self._grid.active_links] = (
            self._grid.calc_grad_at_link(self._wtable) * cosa
        )[self._grid.active_links]

        # Calculate groundwater velocity
        self._vel[:] = -self._K * self._hydr_grad

        # Aquifer thickness at links (upwind)
        hlink = (
            map_value_at_max_node_to_link(
                self._grid, "water_table__elevation", "aquifer__thickness"
            )
            * cosa
        )

        # Calculate specific discharge
        self._q[:] = hlink * self._vel

        # Groundwater flux divergence
        dqdx = self._grid.calc_flux_div_at_node(self._q)

        # Regolith thickness
        reg_thickness = self._elev - self._base

        # update thickness from analytical
        self._thickness[self._cores] = _update_thickness(
            dt, self._thickness, reg_thickness, self._recharge, dqdx, self._n, self._r
        )[self._cores]
        self._thickness[self._thickness < 0] = 0.0

        # Recalculate water surface height
        self._wtable[:] = self._base + self._thickness

        # Calculate surface discharge at nodes
        self._qs[:] = _regularize_G(
            self._thickness / reg_thickness, self._r
        ) * _regularize_R(self._recharge - dqdx)

    def run_with_adaptive_time_step_solver(self, dt):
        """
        Advance component by one time step of size dt, subdividing the timestep
        into substeps as necessary to meet stability conditions.
        Note this method returns the fluxes at the last substep, but also
        returns a new field, average_surface_water__specific_discharge, that is
        averaged over all subtimesteps. To return state during substeps,
        provide a callback_fun.

        Parameters
        ----------
        dt: float
            The imposed timestep.
        """

        # check water table above surface
        if (self._wtable > self._elev).any():
            warn(
                "water table above elevation surface. "
                "Setting water table elevation here to "
                "elevation surface"
            )
            self._wtable[self._wtable > self._elev] = self._elev[
                self._wtable > self._elev
            ]
            self._thickness[self._cores] = (self._wtable - self._base)[self._cores]

        # Calculate base gradient
        self._base_grad[self._grid.active_links] = self._grid.calc_grad_at_link(
            self._base
        )[self._grid.active_links]
        cosa = np.cos(np.arctan(self._base_grad))

        # Initialize reg_thickness
        reg_thickness = self._elev - self._base

        # Initialize for average surface discharge
        qs_cumulative = np.zeros_like(self._elev)

        # Initialize variable timestep
        remaining_time = dt
        self._num_substeps = 0

        while remaining_time > 0.0:

            # Calculate hydraulic gradient
            self._hydr_grad[self._grid.active_links] = (
                self._grid.calc_grad_at_link(self._wtable) * cosa
            )[self._grid.active_links]

            # Calculate groundwater velocity
            self._vel[:] = -self._K * self._hydr_grad

            # Aquifer thickness at links (upwind)
            hlink = (
                map_value_at_max_node_to_link(
                    self._grid, "water_table__elevation", "aquifer__thickness"
                )
                * cosa
            )

            # Calculate specific discharge
            self._q[:] = hlink * self._vel

            # Groundwater flux divergence
            dqdx = self._grid.calc_flux_div_at_node(self._q)

            # calculate criteria for timestep
            dt_vn = self._vn_coefficient * np.min(
                np.divide(
                    (self._n_link * self._grid.length_of_link ** 2),
                    (4 * self._K * hlink),
                    where=hlink > 0,
                    out=np.ones_like(self._q) * 1e15,
                )
            )

            dt_courant = self._courant_coefficient * np.min(
                np.divide(
                    self._grid.length_of_link,
                    abs(self._vel / self._n_link),
                    where=abs(self._vel) > 0,
                    out=np.ones_like(self._q) * 1e15,
                )
            )
            substep_dt = min([dt_courant, dt_vn, remaining_time])
            # print(np.argmin(np.array([self._dt_courant, self._dt_vn, remaining_time]))) # 0 = courant limited, 1 = vn limited, 2 = not limited

            # update thickness from analytical
            self._thickness[self._cores] = _update_thickness(
                substep_dt,
                self._thickness,
                reg_thickness,
                self._recharge,
                dqdx,
                self._n,
                self._r,
            )[self._cores]
            self._thickness[self._thickness < 0] = 0.0

            # Recalculate water surface height
            self._wtable[:] = self._base + self._thickness

            # Calculate surface discharge at nodes
            self._qs[:] = _regularize_G(
                self._thickness / reg_thickness, self._r
            ) * _regularize_R(self._recharge - dqdx)

            # add cumulative sw discharge in substeps
            qs_cumulative += self._qs * substep_dt

            # calculate the time remaining and advance count of substeps
            remaining_time -= substep_dt
            self._num_substeps += 1

            # run callback function if supplied
            self._callback_fun(
                self._grid, self._recharge, substep_dt, **self._callback_kwds
            )

        self._qsavg[:] = qs_cumulative / dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  5 08:42:24 2021

@author: gtucker
"""

from numpy.testing import assert_array_almost_equal, assert_array_equal, assert_raises

from landlab import HexModelGrid, RadialModelGrid, RasterModelGrid
from landlab.components import Flexure, ListricKinematicExtender


def test_hangingwall_nodes():
    """Test the correct identification of hangingwall nodes."""
    grid = RasterModelGrid((3, 7), xy_spacing=2500.0)
    grid.add_zeros("topographic__elevation", at="node")
    extender = ListricKinematicExtender(grid, fault_location=2500.0)

    assert_array_equal(
        extender._hangwall, [2, 3, 4, 5, 6, 9, 10, 11, 12, 13, 16, 17, 18, 19, 20]
    )


def test_subsidence_and_horiz_shift():
    """Test that elev subsides then shifts after 2 time steps."""
    grid = RasterModelGrid((3, 7), xy_spacing=2500.0)
    topo = grid.add_zeros("topographic__elevation", at="node")
    extender = ListricKinematicExtender(
        grid, extension_rate=0.01, fault_location=2500.0
    )

    # Run long enough to extend by half a grid cell
    extender.run_one_step(dt=125000.0)
    assert_array_almost_equal(
        topo[7:14],
        [0.0, 0.0, -1404.156819, -910.66907, -590.616478, -383.045648, -248.425118],
    )

    # Now extend another half cell, so cumulative extension is one cell and
    # elevations should get shifted by one cell
    extender.run_one_step(dt=125000.0)
    assert_array_almost_equal(
        topo[7:14],
        [0.0, 0.0, -3514.477461, -2808.313638, -1821.338140, -1181.232956, -766.091296],
    )

    # Another step, and this time the hangingwall edge has moved by one cell,
    # so the first 3 cells in this row should not further subside.
    extender.run_one_step(dt=125000.0)
    assert_array_almost_equal(
        topo[7:14],
        [
            0.0,
            0.0,
            -3514.477461,
            -3718.982708,
            -2411.954617,
            -1564.278603,
            -1014.516414,
        ],
    )


def test_with_hex_grid():
    grid = HexModelGrid((5, 5), node_layout="rect")
    grid.add_zeros("topographic__elevation", at="node")
    ListricKinematicExtender(grid)
    ListricKinematicExtender(grid, fault_location=2.0)

    grid = HexModelGrid((5, 5), node_layout="rect", orientation="vertical")
    grid.add_zeros("topographic__elevation", at="node")
    assert_raises(NotImplementedError, ListricKinematicExtender, grid)


def test_with_flexure():
    """Test integrating with flexure."""
    crust_density = 2700.0  # density of crustal column, kg/m3
    dx = 2500.0  # grid spacing, m
    dt = 125000.0  # time step, y
    upper_crust_base_depth = 10000.0  # m

    grid = RasterModelGrid((3, 7), xy_spacing=dx)
    topo = grid.add_zeros("topographic__elevation", at="node")
    load = grid.add_zeros("lithosphere__overlying_pressure_increment", at="node")
    thickness = grid.add_zeros("upper_crust_thickness", at="node")
    upper_crust_base = grid.add_zeros("upper_crust_base__elevation", at="node")

    extender = ListricKinematicExtender(
        grid,
        extension_rate=0.01,
        fault_location=2500.0,
        track_crustal_thickness=True,
    )
    flexer = Flexure(grid, eet=5000.0, method="flexure")
    deflection = grid.at_node["lithosphere_surface__elevation_increment"]

    topo[
        grid.x_of_node <= 7500.0
    ] = 1000.0  # this will force thickness to be 1 km greater at left
    upper_crust_base[:] = -upper_crust_base_depth
    thickness[:] = topo - upper_crust_base
    unit_wt = crust_density * flexer.gravity
    load[:] = unit_wt * thickness  # loading pressure

    # Get the initial deflection, which we'll need to calculate total current
    # deflection
    flexer.update()
    init_deflection = deflection.copy()

    # Run extension for half a grid cell. Elevations change, but thickness
    # doesn't, so deflection should not change. We should be able to recover
    # elevation from:
    #
    #   topo = thickness + crust base - (deflection + subsidence)
    #
    extender.run_one_step(dt=dt)
    flexer.update()
    net_deflection = deflection - init_deflection
    assert_array_almost_equal(
        net_deflection[7:14],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    )
    test_topo = thickness + upper_crust_base - (net_deflection + extender._cum_subs)
    assert_array_almost_equal(topo, test_topo)

    # Now extend for another half cell, which should force a shift. The
    # cumulative subsidence will be subtracted from the thickness field,
    # representing thinning as the hangingwall slides to the "right". This
    # will cause net upward isostatic deflection.
    extender.run_one_step(dt=dt)
    load[:] = unit_wt * thickness
    flexer.update()
    net_deflection = deflection - init_deflection
    assert_array_almost_equal(
        thickness[7:14],
        [
            11000.0,
            11000.0,
            8191.686362,  # greatest subsidence: lost nearly 3 km
            9178.66186,
            9818.767044,  # thicker because shifted (only lost <200 m)
            9233.908704,
            9503.149763,
        ],
    )
    assert_array_almost_equal(
        net_deflection[7:14],
        [
            -59.497362,
            -65.176276,
            -69.222531,
            -70.334462,
            -68.608952,
            -64.912352,
            -59.743080,
        ],
    )


def test_error_handling():

    radial_grid = RadialModelGrid(
        n_rings=1, nodes_in_first_ring=8
    )  # , xy_of_center=(0., 0.))
    assert_raises(TypeError, ListricKinematicExtender, radial_grid)

    hex_grid = HexModelGrid((3, 3))
    assert_raises(TypeError, ListricKinematicExtender, hex_grid)

    grid = RasterModelGrid((3, 7))
    grid.add_zeros("topographic__elevation", at="node")
    assert_raises(
        KeyError, ListricKinematicExtender, grid, track_crustal_thickness=True
    )



import pytest

from landlab import RasterModelGrid
from landlab.layers import MaterialLayers


def test_MaterialLayersMixIn():
    grid = RasterModelGrid((4, 4))
    assert hasattr(grid, "material_layers")
    assert grid.material_layers.number_of_layers == 0
    assert grid.material_layers.number_of_stacks == 4


def test_adding_untracked_layer():
    layers = MaterialLayers(3)
    layers.add(1.0, type=3.0, size="sand")
    layers.add([0.0, 0.0, 1.0], type=3.0, size="sand")
    with pytest.raises(ValueError):
        layers.add([1.0], type=3.0, size="sand", spam="eggs")
